#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import os

from typing import TYPE_CHECKING

import pytest

from molecule import config
from molecule.driver import delegated
from tests.conftest import is_subset  # pylint:disable=C0411


if TYPE_CHECKING:
    from pytest_mock import MockerFixture


@pytest.fixture
def _driver_managed_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "driver": {
            "name": "default",
            "options": {
                "managed": True,
            },
        },
    }


@pytest.fixture
def _driver_unmanaged_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "driver": {
            "name": "default",
            "options": {
                "managed": False,
            },
        },
    }


@pytest.fixture
def _instance(config_instance: config.Config):  # type: ignore[no-untyped-def]  # noqa: ANN202
    return delegated.Delegated(config_instance)


def test_delegated_config_private_member(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert isinstance(_instance._config, config.Config)


def test_delegated_options_property2(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.testinfra_options == {
        "connection": "ansible",
        "ansible-inventory": _instance._config.provisioner.inventory_directory,
    }


def test_delegated_name_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.name == "default"


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_unmanaged_section_data"],  # noqa: PT007
    indirect=True,
)
def test_delegated_options_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {
        "managed": False,
    }

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_managed_section_data"],  # noqa: PT007
    indirect=True,
)
def test_delegated_options_property_when_managed(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {"managed": True}

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_managed_section_data"],  # noqa: PT007
    indirect=True,
)
def test_login_cmd_template_property_when_managed(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = (
        "ssh {address} -l {user} -p {port} -i {identity_file} "
        "-o UserKnownHostsFile=/dev/null "
        "-o ControlMaster=auto "
        "-o ControlPersist=60s "
        "-o ForwardX11=no "
        "-o LogLevel=ERROR "
        "-o IdentitiesOnly=yes "
        "-o StrictHostKeyChecking=no"
    )

    assert x == _instance.login_cmd_template


def test_safe_files_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.safe_files == []


def test_default_safe_files_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.default_safe_files == []


def test_delegated_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.delegated


def test_managed_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.managed


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_unmanaged_section_data"],  # noqa: PT007
    indirect=True,
)
def test_default_ssh_connection_options_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.default_ssh_connection_options == []


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_managed_section_data"],  # noqa: PT007
    indirect=True,
)
def test_default_ssh_connection_options_property_when_managed(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = [
        "-o UserKnownHostsFile=/dev/null",
        "-o ControlMaster=auto",
        "-o ControlPersist=60s",
        "-o ForwardX11=no",
        "-o LogLevel=ERROR",
        "-o IdentitiesOnly=yes",
        "-o StrictHostKeyChecking=no",
    ]

    assert x == _instance.default_ssh_connection_options


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_unmanaged_section_data"],  # noqa: PT007
    indirect=True,
)
def test_login_options(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.login_options("foo") == {"instance": "foo"}


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_managed_section_data"],  # noqa: PT007
    indirect=True,
)
def test_login_options_when_managed(mocker: MockerFixture, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    m = mocker.patch("molecule.driver.delegated.Delegated._get_instance_config")
    m.return_value = {
        "instance": "foo",
        "address": "172.16.0.2",
        "user": "cloud-user",
        "port": 22,
        "become_method": "su",
        "become_pass": "password",
        "identity_file": "/foo/bar",
    }

    x = {
        "instance": "foo",
        "address": "172.16.0.2",
        "user": "cloud-user",
        "port": 22,
        "become_method": "su",
        "become_pass": "password",
        "identity_file": "/foo/bar",
    }
    assert x == _instance.login_options("foo")


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_unmanaged_section_data"],  # noqa: PT007
    indirect=True,
)
def test_ansible_connection_options(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {}  # type: ignore[var-annotated]

    assert is_subset(x, _instance.ansible_connection_options("foo"))  # type: ignore[no-untyped-call]


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_managed_section_data"],  # noqa: PT007
    indirect=True,
)
def test_ansible_connection_options_when_managed(mocker: MockerFixture, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.managed is True

    ssh_case_data = mocker.patch(
        "molecule.driver.delegated.Delegated._get_instance_config",
    )
    ssh_case_data.return_value = {
        "instance": "foo",
        "address": "172.16.0.2",
        "user": "cloud-user",
        "port": 22,
        "become_method": "su",
        "become_pass": "password",
        "identity_file": "/foo/bar",
    }

    ssh_expected_data = {
        "ansible_host": "172.16.0.2",
        "ansible_port": 22,
        "ansible_user": "cloud-user",
        "ansible_become_method": "su",
        "ansible_become_pass": "password",
        "ansible_private_key_file": "/foo/bar",
        "ansible_connection": "smart",
        "ansible_ssh_common_args": (
            "-o UserKnownHostsFile=/dev/null "
            "-o ControlMaster=auto "
            "-o ControlPersist=60s "
            "-o ForwardX11=no "
            "-o LogLevel=ERROR "
            "-o IdentitiesOnly=yes "
            "-o StrictHostKeyChecking=no"
        ),
    }

    assert ssh_expected_data.items() <= _instance.ansible_connection_options("foo").items()

    winrm_case_data = mocker.patch(
        "molecule.driver.delegated.Delegated._get_instance_config",
    )
    winrm_case_data.return_value = {
        "instance": "foo",
        "address": "172.16.0.2",
        "user": "cloud-user",
        "port": 5896,
        "connection": "winrm",
    }

    winrm_expected_data = {
        "ansible_host": "172.16.0.2",
        "ansible_port": 5896,
        "ansible_user": "cloud-user",
        "ansible_connection": "winrm",
    }

    assert winrm_expected_data.items() <= _instance.ansible_connection_options("foo").items()


def test_ansible_connection_options_handles_missing_instance_config_managed(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    mocker: MockerFixture,
    _instance,  # noqa: PT019
):
    m = mocker.patch("molecule.util.safe_load_file")
    m.side_effect = IOError

    assert _instance.ansible_connection_options("foo") == {}


def test_ansible_connection_options_handles_missing_results_key_when_managed(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    mocker: MockerFixture,
    _instance,  # noqa: PT019
):
    m = mocker.patch("molecule.util.safe_load_file")
    m.side_effect = StopIteration

    assert _instance.ansible_connection_options("foo") == {}


def test_instance_config_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = os.path.join(  # noqa: PTH118
        _instance._config.scenario.ephemeral_directory,
        "instance_config.yml",
    )

    assert x == _instance.instance_config


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_unmanaged_section_data"],  # noqa: PT007
    indirect=True,
)
def test_ssh_connection_options_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.ssh_connection_options == []


def test_status(mocker: MockerFixture, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    result = _instance.status()

    assert len(result) == 2  # noqa: PLR2004

    assert result[0].instance_name == "instance-1"
    assert result[0].driver_name == "default"
    assert result[0].provisioner_name == "ansible"
    assert result[0].scenario_name == "default"
    assert result[0].created == "false"
    assert result[0].converged == "false"

    assert result[1].instance_name == "instance-2"
    assert result[1].driver_name == "default"
    assert result[1].provisioner_name == "ansible"
    assert result[1].scenario_name == "default"
    assert result[1].created == "false"
    assert result[1].converged == "false"


def test_delegated_created(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._created() == "false"


@pytest.fixture
def _driver_options_managed_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"driver": {"options": {"managed": False}}}


@pytest.mark.parametrize(
    "config_instance",
    ["_driver_options_managed_section_data"],  # noqa: PT007
    indirect=True,
)
def test_created_unknown_when_managed_false(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    _driver_options_managed_section_data,  # noqa: PT019
    _instance,  # noqa: PT019
):
    assert _instance._created() == "unknown"


def test_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._converged() == "false"


def test_get_instance_config(mocker: MockerFixture, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    m = mocker.patch("molecule.util.safe_load_file")
    m.return_value = [{"instance": "foo"}, {"instance": "bar"}]

    x = {"instance": "foo"}
    assert x == _instance._get_instance_config("foo")



"""Molecule API Module."""

from __future__ import annotations

import logging
import traceback

from collections import OrderedDict
from typing import TYPE_CHECKING

import pluggy

from ansible_compat.ports import cache

from molecule.driver.base import Driver
from molecule.exceptions import MoleculeError
from molecule.verifier.base import Verifier


if TYPE_CHECKING:
    from molecule.config import Config


LOG = logging.getLogger(__name__)


class MoleculeRuntimeWarning(RuntimeWarning):
    """A runtime warning used by Molecule and its plugins."""


class IncompatibleMoleculeRuntimeWarning(MoleculeRuntimeWarning):
    """A warning noting an unsupported runtime environment."""


@cache
def drivers(config: Config | None = None) -> dict[str, Driver]:
    """Return list of active drivers.

    Args:
        config: plugin config

    Returns:
        A dictionary of active drivers by name.
    """
    plugins = OrderedDict()
    pm = pluggy.PluginManager("molecule.driver")
    try:
        pm.load_setuptools_entrypoints("molecule.driver")
    except MoleculeError:
        # These are not fatal because a broken driver should not make the entire
        # tool unusable.
        LOG.error("Failed to load driver entry point %s", traceback.format_exc())  # noqa: TRY400
    for plugin in pm.get_plugins():
        if issubclass(plugin, Driver):
            try:
                driver = plugin(config)
                plugins[driver.name] = driver
            except (MoleculeError, TypeError) as e:
                LOG.error(  # noqa: TRY400
                    "Failed to load %s driver: %s",
                    pm.get_name(plugin),
                    str(e),
                )
        else:
            msg = f"Skipped loading plugin class {plugin} because is not a subclass of Driver."
            LOG.error(msg)

    return plugins


@cache
def verifiers(config: Config | None = None) -> dict[str, Verifier]:
    """Return list of active verifiers.

    Args:
        config: plugin config

    Returns:
        A dictionary of active verifiers by name.
    """
    plugins = OrderedDict()
    pm = pluggy.PluginManager("molecule.verifier")
    try:
        pm.load_setuptools_entrypoints("molecule.verifier")
    except Exception:  # noqa: BLE001
        # These are not fatal because a broken verifier should not make the entire
        # tool unusable.
        LOG.error("Failed to load verifier entry point %s", traceback.format_exc())  # noqa: TRY400
    for plugin_class in pm.get_plugins():
        try:
            if issubclass(plugin_class, Verifier):
                plugin = plugin_class(config)
                plugins[plugin.name] = plugin
        except Exception as e:  # noqa: BLE001, PERF203
            LOG.error("Failed to load %s driver: %s", plugin_class.__name__, str(e))  # noqa: TRY400
    return plugins


__all__ = (
    "Driver",
    "IncompatibleMoleculeRuntimeWarning",
    "MoleculeRuntimeWarning",
    "Verifier",
    "drivers",
    "verifiers",
)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Verifier Base Module."""

from __future__ import annotations

import abc

from pathlib import Path
from typing import TYPE_CHECKING, TypedDict

from molecule import util


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from molecule.config import Config


class VerifierSchemaName(TypedDict):
    """Verifier schema name definition.

    Attributes:
        type: schema type.
        allowed: list of allowed verifier names.
    """

    type: str
    allowed: list[str]


class VerifierSchema(TypedDict):
    """Verifier schema.

    Attributes:
        name: Schema name container.
    """

    name: VerifierSchemaName


class VerifierDef(TypedDict):
    """Verifier schema container.

    Attributes:
        type: Schema type.
        schema: Schema container.
    """

    type: str
    schema: VerifierSchema


class Schema(TypedDict):
    """Verifier schema definition.

    Attributes:
        verifier: Verifier schema container.
    """

    verifier: VerifierDef


class Verifier(abc.ABC):
    """Verifier Base Class."""

    def __init__(self, config: Config) -> None:
        """Initialize code for all :ref:`Verifier` classes.

        Args:
            config: An instance of a Molecule config.
        """
        self._config = config

    @property
    @abc.abstractmethod
    def name(self) -> str:  # pragma: no cover
        """Name of the verifier.

        Returns:
            The name of the verifier.
        """

    @property
    @abc.abstractmethod
    def default_options(self) -> MutableMapping[str, str | bool]:  # pragma: no cover
        """Get default CLI arguments provided to ``cmd``.

        Returns:
            The default verifier options.
        """

    @property
    @abc.abstractmethod
    def default_env(self) -> dict[str, str]:  # pragma: no cover
        """Get default env variables provided to ``cmd``.

        Returns:
            The default verifier environment variables.
        """

    @abc.abstractmethod
    def execute(self, action_args: list[str] | None = None) -> None:  # pragma: no cover
        """Execute ``cmd``.

        Args:
            action_args: list of arguments to be passed.
        """

    @abc.abstractmethod
    def schema(self) -> Schema:  # pragma: no cover
        """Return validation schema.

        Returns:
            Verifier schema.
        """

    @property
    def enabled(self) -> bool:
        """Is the verifier enabled.

        Returns:
            Whether the verifier is enabled.
        """
        return self._config.config["verifier"]["enabled"]

    @property
    def directory(self) -> str:
        """Verifier directory.

        Returns:
            The directory of the the verifier files.
        """
        return str(
            Path(
                self._config.scenario.directory,
                self._config.config["verifier"].get("directory", "tests"),
            ),
        )

    @property
    def options(self) -> MutableMapping[str, str | bool]:
        """The computed options for this verifier.

        Returns:
            The combined dictionary of default options and those specified in the config.
        """
        return util.merge_dicts(
            self.default_options,
            self._config.config["verifier"]["options"],
        )

    @property
    def env(self) -> dict[str, str]:
        """The computed environment variables for this verifier.

        Returns:
            The combined dictionary of default environment variables and those
            specified in the config.
        """
        return util.merge_dicts(
            self.default_env,
            self._config.config["verifier"]["env"],
        )

    def __eq__(self, other: object) -> bool:
        """Implement equality comparison.

        Args:
            other: object to compare against this one.

        Returns:
            Whether other matches the name of this verifier.
        """
        return str(self) == str(other)

    def __lt__(self, other: object) -> bool:
        """Implement lower than comparison.

        Args:
            other: object to compare against this one.

        Returns:
            Whether other is less than the name of this verifier.
        """
        return str.__lt__(str(self), str(other))

    def __hash__(self) -> int:
        """Implement hashing.

        Returns:
            The hash of the verifier's name.
        """
        return self.name.__hash__()

    def __str__(self) -> str:
        """Return readable string representation of object.

        Returns:
            The verifier's name.
        """
        return self.name

    def __repr__(self) -> str:
        """Return detailed string representation of object.

        Returns:
            The verifier's name.
        """
        return self.name



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import logging

from typing import TYPE_CHECKING

import pytest

from molecule.ansi_output import should_do_markup
from molecule.command.base import Base
from molecule.logger import get_scenario_logger, get_section_loggers


if TYPE_CHECKING:
    from molecule import config


# the dummy/instance fixtures are based on fixtures in tests.unit.command.test_base
class Dummy(Base):
    """ExtendedBase Class."""

    def execute(self, action_args=None):  # type: ignore[no-untyped-def]  # noqa: ANN201, ARG002, D102
        return True


@pytest.fixture
def _dummy_class(patched_config_validate, config_instance: config.Config):  # type: ignore[no-untyped-def]  # noqa: ANN202
    return Dummy


@pytest.fixture
def _instance(_dummy_class, config_instance, _patched_logger_env):  # type: ignore[no-untyped-def]  # noqa: ANN202
    # _patched_logger_env included here to ensure pytest runs it first
    get_section_loggers.cache_clear()
    return _dummy_class(config_instance)


@pytest.fixture
def _patched_logger_env(request, monkeypatch):  # type: ignore[no-untyped-def]  # noqa: ANN202
    """Parametrize tests with and without CI env vars."""
    envvars = {"CI": None, "GITHUB_ACTIONS": None, "GITLAB_CI": None, "TRAVIS": None}
    envvars.update(request.param[1])
    for envvar, value in envvars.items():
        if value is None:
            monkeypatch.delenv(envvar, raising=False)
        else:
            monkeypatch.setenv(envvar, value)
    return request.param[0]


get_section_logger_tests = [
    # (expected # of section_loggers, envvars)
    (1, {}),
    (2, {"CI": "true", "GITHUB_ACTIONS": "true"}),
    (2, {"CI": "true", "GITLAB_CI": "true"}),
    (2, {"CI": "true", "TRAVIS": "true"}),
    (1, {"CI": "true", "RANDOM_CI": "true"}),
]


@pytest.mark.parametrize(
    "_patched_logger_env",
    get_section_logger_tests,
    indirect=True,
)
def test_get_section_loggers(_patched_logger_env):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    expected_section_loggers = _patched_logger_env
    get_section_loggers.cache_clear()
    section_loggers = get_section_loggers()
    assert len(section_loggers) == expected_section_loggers  # type: ignore[arg-type]


@pytest.mark.parametrize(
    "_patched_logger_env",
    get_section_logger_tests,
    indirect=True,
)
def test_section_loggers_do_not_change_behavior(_patched_logger_env, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    dummy_return = _instance.execute()
    assert dummy_return is True


def test_markup_detection_pycolors0(monkeypatch):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    monkeypatch.setenv("PY_COLORS", "0")
    assert not should_do_markup()


def test_markup_detection_pycolors1(monkeypatch):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setenv("PY_COLORS", "1")
    assert should_do_markup()


def test_markup_detection_tty_yes(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    mocker.patch("sys.stdout.isatty", return_value=True)
    mocker.patch("os.environ", {"TERM": "xterm"})
    assert should_do_markup()
    mocker.resetall()
    mocker.stopall()


def test_markup_detection_tty_no(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    mocker.patch("os.environ", {})
    mocker.patch("sys.stdout.isatty", return_value=False)
    assert not should_do_markup()
    mocker.resetall()
    mocker.stopall()


def test_logger_class():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    class FooLogger(logging.getLoggerClass()):  # type: ignore[misc]
        """stub logger that subclasses logging.getLoggerClass()."""

    logging.setLoggerClass(FooLogger)

    # this test throws RecursionError prior to bugfix
    assert FooLogger("foo")


def test_scenario_logger_with_step(caplog):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    # Test logger with step
    logger_with_step = get_scenario_logger("test", "test_scenario", "converge")

    with caplog.at_level(logging.INFO):
        logger_with_step.info("Test message")

    # Check that the log record has both scenario and step information
    record = caplog.records[0]
    assert hasattr(record, "molecule_scenario")
    assert hasattr(record, "molecule_step")
    assert record.molecule_scenario == "test_scenario"
    assert record.molecule_step == "converge"

    # Test logger with step parameter
    caplog.clear()
    logger_without_step = get_scenario_logger("test", "test_scenario", "test")

    with caplog.at_level(logging.INFO):
        logger_without_step.info("Test message")

    # Check that the log record has both scenario and step information
    record = caplog.records[0]
    assert hasattr(record, "molecule_scenario")
    assert record.molecule_step == "test"
    assert record.molecule_scenario == "test_scenario"



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Base Command Module."""

from __future__ import annotations

import abc
import collections
import contextlib
import copy
import importlib
import logging
import shutil
import subprocess

from typing import TYPE_CHECKING, Any

import wcmatch.pathlib
import wcmatch.wcmatch

from wcmatch import glob

from molecule import config, logger, text, util
from molecule.console import console
from molecule.constants import MOLECULE_DEFAULT_SCENARIO_NAME, MOLECULE_GLOB
from molecule.exceptions import MoleculeError, ScenarioFailureError
from molecule.scenarios import Scenarios
from molecule.util import safe_dump


if TYPE_CHECKING:
    from molecule.scenario import Scenario
    from molecule.types import CommandArgs, MoleculeArgs, ScenariosResults


def _log(scenario_name: str, step: str, message: str, level: str = "info") -> None:
    """Create scenario logger on-demand and log message.

    Args:
        scenario_name: Name of the scenario for context.
        step: Step name for context (e.g., 'discovery', 'prerun', 'reset').
        message: Log message (pre-formatted, no placeholders).
        level: Log level ('info', 'warning', 'error', 'debug').
    """
    scenario_log = logger.get_scenario_logger(__name__, scenario_name, step)
    getattr(scenario_log, level)(message)


class Base(abc.ABC):
    """An abstract base class used to define the command interface."""

    def __init__(self, c: config.Config) -> None:
        """Initialize code for all command classes.

        Args:
            c: An instance of a Molecule config.
        """
        self._config = c
        self._setup()

    def __init_subclass__(cls) -> None:
        """Decorate execute from all subclasses."""
        super().__init_subclass__()
        for wrapper in logger.get_section_loggers():
            cls.execute = wrapper(cls.execute)  # type: ignore[method-assign,assignment]

    @property
    def _log(self) -> logger.ScenarioLoggerAdapter:
        """Get a scenario logger with command-specific step context.

        Returns:
            A scenario logger adapter with current scenario and step context.
        """
        step_name = self.__class__.__name__.lower()
        return logger.get_scenario_logger(__name__, self._config.scenario.name, step_name)

    @abc.abstractmethod
    def execute(
        self,
        action_args: list[str] | None = None,
    ) -> None:  # pragma: no cover
        """Abstract method to execute the command.

        Args:
            action_args: An optional list of arguments to pass to the action.
        """

    def _setup(self) -> None:
        """Prepare Molecule's provisioner and returns None."""
        self._config.write()
        if self._config.provisioner is not None:
            self._config.provisioner.write_config()
            self._config.provisioner.manage_inventory()


def execute_cmdline_scenarios(
    scenario_names: list[str] | None,
    args: MoleculeArgs,
    command_args: CommandArgs,
    ansible_args: tuple[str, ...] = (),
    excludes: list[str] | None = None,
) -> None:
    """Execute scenario sequences based on parsed command-line arguments.

    This is useful for subcommands that run scenario sequences, which
    excludes subcommands such as ``list``, ``login``, and ``matrix``.

    ``args`` and ``command_args`` are combined using :func:`get_configs`
    to generate the scenario(s) configuration.

    Args:
        scenario_names: Name of scenarios to run, or ``None`` to run all.
        args: ``args`` dict from ``click`` command context
        command_args: dict of command arguments, including the target
        ansible_args: Optional tuple of arguments to pass to the `ansible-playbook` command
        excludes: Name of scenarios to not run.
    """
    if excludes is None:
        excludes = []

    configs: list[config.Config] = []
    if scenario_names is None:
        configs = [
            config
            for config in get_configs(args, command_args, ansible_args, MOLECULE_GLOB)
            if config.scenario.name not in excludes
        ]
    else:
        try:
            # filter out excludes
            scenario_names = [name for name in scenario_names if name not in excludes]
            for scenario_name in scenario_names:
                glob_str = MOLECULE_GLOB.replace("*", scenario_name)
                configs.extend(get_configs(args, command_args, ansible_args, glob_str))
        except ScenarioFailureError as exc:
            util.sysexit(code=exc.code)

    default_glob = MOLECULE_GLOB.replace("*", MOLECULE_DEFAULT_SCENARIO_NAME)
    default_config = None
    try:
        default_config = get_configs(args, command_args, ansible_args, default_glob)[0]
    except MoleculeError:
        # Use a generic logger for this since it's not tied to a specific scenario
        logging.getLogger(__name__).info("default scenario not found, disabling shared state.")

    scenarios = _generate_scenarios(scenario_names, configs)

    try:
        _run_scenarios(scenarios, command_args, default_config)

    except ScenarioFailureError as exc:
        util.sysexit(code=exc.code)
    finally:
        if command_args.get("report"):
            console.print(generate_report(scenarios.results))


def _generate_scenarios(
    scenario_names: list[str] | None,
    configs: list[config.Config],
) -> Scenarios:
    """Generate Scenarios object from names and configs.

    Args:
        scenario_names: Names of scenarios to include.
        configs: List of Config objects to consider.

    Returns:
        Combined Scenarios object.
    """
    scenarios = Scenarios(
        configs,
        scenario_names,
    )

    if scenario_names is not None:
        for scenario_name in scenario_names:
            if scenario_name != "*" and scenarios:
                # Use generic "discovery" step since this is scenario discovery phase
                _log(
                    scenario_name,
                    "discovery",
                    f"scenario test matrix: {', '.join(scenarios.sequence(scenario_name))}",
                )

    return scenarios


def _run_scenarios(
    scenarios: Scenarios,
    command_args: CommandArgs,
    default_config: config.Config | None,
) -> None:
    """Loop through Scenarios object and execute each.

    Args:
        scenarios: The Scenarios object holding all of the Scenario objects.
        command_args: dict of command arguments.
        default_config: Molecule Config object for the default scenario.

    Raises:
        ScenarioFailureError: when a scenario fails prematurely.
    """
    # Run initial create
    create_results = execute_subcommand_default(default_config, "create")
    if create_results is not None:
        scenarios.results.append(create_results)

    for scenario in scenarios.all:
        if scenario.config.config["prerun"]:
            role_name_check = scenario.config.config["role_name_check"]
            _log(
                scenario.config.scenario.name,
                "prerun",
                f"Performing prerun with role_name_check={role_name_check}...",
            )
            scenario.config.runtime.prepare_environment(
                install_local=True,
                role_name_check=role_name_check,
            )

        if command_args.get("subcommand") == "reset":
            _log(
                scenario.config.scenario.name,
                "reset",
                f"Removing {scenario.ephemeral_directory}",
            )
            shutil.rmtree(scenario.ephemeral_directory)
            return
        try:
            execute_scenario(scenario)
            scenarios.results.append({"name": scenario.name, "results": scenario.results})
        except ScenarioFailureError:
            # if the command has a 'destroy' arg, like test does,
            # handle that behavior here.
            if command_args.get("destroy") == "always":
                msg = (
                    f"An error occurred during the {scenario.config.subcommand} sequence action: "
                    f"'{scenario.config.action}'. Cleaning up."
                )
                step_name = getattr(scenario.config, "action", "cleanup")
                _log(
                    scenario.config.scenario.name,
                    step_name,
                    msg,
                    level="warning",
                )
                execute_subcommand(scenario.config, "cleanup")
                destroy_results = execute_subcommand_default(default_config, "destroy")
                if destroy_results is not None:
                    scenarios.results.append({"name": scenario.name, "results": scenario.results})
                    scenarios.results.append(destroy_results)
                else:
                    execute_subcommand(scenario.config, "destroy")
                    scenarios.results.append({"name": scenario.name, "results": scenario.results})

                # always prune ephemeral dir if destroying on failure
                scenario.prune()
                if scenario.config.is_parallel:
                    scenario._remove_scenario_state_directory()  # noqa: SLF001
            raise

    # Run final destroy
    destroy_results = execute_subcommand_default(default_config, "destroy")
    if destroy_results is not None:
        scenarios.results.append(destroy_results)


def execute_subcommand_default(
    default_config: config.Config | None,
    subcommand: str,
) -> ScenariosResults | None:
    """Execute subcommand as in execute_subcommand, but do it from the default scenario if one exists.

    Args:
        default_config: The Config object for the default scenario, if it exists.
        subcommand: The desired subcommand to run.

    Returns:
        The result of the subcommand.
    """
    if default_config is None or default_config.shared_data is not True:
        # We have not been asked to do anything.
        return None

    default = default_config.scenario
    if subcommand in default.sequence:
        execute_subcommand(default_config, subcommand)
        results: ScenariosResults = {"name": default.name, "results": copy.copy(default.results)}
        # clear results for later reuse
        default.results = []
        return results
    _log(
        default.name,
        subcommand,
        f"{subcommand} not found in default scenario, falling back to current scenario",
        level="warning",
    )
    return None


def execute_subcommand(
    current_config: config.Config,
    subcommand_and_args: str,
) -> Any:  # noqa: ANN401
    """Execute subcommand.

    Args:
        current_config: An instance of a Molecule config.
        subcommand_and_args: A string representing the subcommand and arguments.

    Returns:
        The result of the subcommand.
    """
    (subcommand, *args) = subcommand_and_args.split(" ")
    command_module = importlib.import_module(f"molecule.command.{subcommand}")
    command = getattr(command_module, text.camelize(subcommand))

    # knowledge of the current action is used by some provisioners
    # to ensure they behave correctly during certain sequence steps,
    # particularly the setting of ansible options in create/destroy,
    # and is also used for reporting in execute_cmdline_scenarios
    current_config.action = subcommand

    return command(current_config).execute(args)


def execute_scenario(scenario: Scenario) -> None:
    """Execute each command in the given scenario's configured sequence.

    Args:
        scenario: The scenario to execute.
    """
    shared_data = scenario.config.shared_data is True
    for action in scenario.sequence:
        if shared_data and action in ("create", "destroy"):
            # Ignore
            continue

        execute_subcommand(scenario.config, action)

    if (
        not shared_data
        and "destroy" in scenario.sequence
        and scenario.config.command_args.get("destroy") != "never"
    ):
        scenario.prune()

        if scenario.config.is_parallel:
            scenario._remove_scenario_state_directory()  # noqa: SLF001


def filter_ignored_scenarios(scenario_paths: list[str]) -> list[str]:
    """Filter out candidate scenario paths that are ignored by git.

    Args:
        scenario_paths: List of candidate scenario paths.

    Returns:
        Filtered list of scenario paths.
    """
    command = ["git", "check-ignore", *scenario_paths]

    with contextlib.suppress(subprocess.CalledProcessError, FileNotFoundError):
        proc = subprocess.run(
            args=command,
            capture_output=True,
            check=True,
            text=True,
            shell=False,
        )

    try:
        ignored = proc.stdout.splitlines()
        paths = [candidate for candidate in scenario_paths if str(candidate) not in ignored]
    except NameError:
        paths = scenario_paths

    return paths


def get_configs(
    args: MoleculeArgs,
    command_args: CommandArgs,
    ansible_args: tuple[str, ...] = (),
    glob_str: str = MOLECULE_GLOB,
) -> list[config.Config]:
    """Glob the current directory for Molecule config files.

    Instantiate config objects, and returns a list.

    Args:
        args: A dict of options, arguments and commands from the CLI.
        command_args: A dict of options passed to the subcommand from the CLI.
        ansible_args: An optional tuple of arguments provided to the `ansible-playbook` command.
        glob_str: A string representing the glob used to find Molecule config files.

    Returns:
        A list of Config objects.
    """
    scenario_paths = glob.glob(
        glob_str,
        flags=wcmatch.pathlib.GLOBSTAR | wcmatch.pathlib.BRACE | wcmatch.pathlib.DOTGLOB,
    )

    scenario_paths = filter_ignored_scenarios(scenario_paths)
    configs = [
        config.Config(
            molecule_file=util.abs_path(c),
            args=args,
            command_args=command_args,
            ansible_args=ansible_args,
        )
        for c in scenario_paths
    ]
    _verify_configs(configs, glob_str)

    return configs


def _verify_configs(configs: list[config.Config], glob_str: str = MOLECULE_GLOB) -> None:
    """Verify a Molecule config was found and returns None.

    Args:
        configs: A list containing absolute paths to Molecule config files.
        glob_str: A string representing the glob used to find Molecule config files.

    Raises:
        ScenarioFailureError: When scenario configs cannot be verified.
    """
    if configs:
        scenario_names = [c.scenario.name for c in configs]
        for scenario_name, n in collections.Counter(scenario_names).items():
            if n > 1:
                msg = f"Duplicate scenario name '{scenario_name}' found.  Exiting."
                raise ScenarioFailureError(message=msg)

    else:
        msg = f"'{glob_str}' glob failed.  Exiting."
        raise ScenarioFailureError(message=msg)


def _get_subcommand(string: str) -> str:
    """Return the subcommand from a string.

    Args:
        string: A string containing a subcommand.

    Returns:
        A string representing the subcommand.
    """
    return string.split(".")[-1]


def generate_report(results: list[ScenariosResults]) -> str:
    """Print end-of-run report.

    Args:
        results: Dictionary containing results from each scenario.

    Returns:
        The formatted end-of-run report.
    """
    return safe_dump(results)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import logging

from typing import TYPE_CHECKING

import pytest

from molecule.command.init import scenario
from molecule.config import Config
from molecule.exceptions import MoleculeError


if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture(name="command_args")
def fixture_command_args() -> dict[str, str]:
    """Provide a dictionary of command arguments."""
    return {
        "driver_name": "default",
        "role_name": "test-role",
        "scenario_name": "test-scenario",
        "subcommand": __name__,
        "verifier_name": "ansible",
    }


@pytest.fixture(name="instance")
def fixture_instance(command_args: scenario.CommandArgs) -> scenario.Scenario:
    """Provide a scenario instance.

    Args:
        command_args: A dictionary of command arguments.
    """
    return scenario.Scenario(Config(""), command_args)


def test_scenario_execute(
    monkeypatch: pytest.MonkeyPatch,
    instance: scenario.Scenario,
    caplog: pytest.LogCaptureFixture,
    tmp_path: Path,
) -> None:
    """Test that the scenario is initialized successfully.

    Args:
        monkeypatch: Pytest fixture.
        instance: A scenario instance.
        caplog: Pytest log capture fixture.
        tmp_path: Pytest temporary directory fixture.
    """
    monkeypatch.chdir(tmp_path)

    # Mock the run_command to avoid ansible execution and create the directory structure
    def mock_run_command(*args: object, **kwargs: object) -> None:
        # Create the molecule/test-scenario directory like ansible-playbook would
        scenario_dir = tmp_path / "molecule" / "test-scenario"
        scenario_dir.mkdir(parents=True, exist_ok=True)

    with monkeypatch.context() as m:
        m.setattr("molecule.app.App.run_command", mock_run_command)

        with caplog.at_level(logging.INFO):
            instance.execute()

    msg = "Initializing new scenario test-scenario..."
    assert any(msg in record.message for record in caplog.records)

    assert (tmp_path / "molecule" / "test-scenario").is_dir()

    scenario_directory = tmp_path / "molecule" / "test-scenario"
    msg = f"Initialized scenario in {scenario_directory} successfully."
    assert any(msg in record.message for record in caplog.records)


def test_execute_scenario_exists(
    monkeypatch: pytest.MonkeyPatch,
    instance: scenario.Scenario,
    caplog: pytest.LogCaptureFixture,
    test_cache_path: Path,
) -> None:
    """Test that the scenario is initialized successfully.

    Args:
        monkeypatch: Pytest fixture.
        instance: A scenario instance.
        caplog: Pytest log capture fixture.
        test_cache_path: Path to the cache directory for the test.
    """
    monkeypatch.chdir(test_cache_path)
    instance.execute()

    with pytest.raises(MoleculeError) as e:
        instance.execute()

    assert e.value.code == 1

    msg = "The directory molecule/test-scenario exists. Cannot create new scenario."
    assert msg in caplog.text



#  Copyright (c) 2019 Red Hat, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Ansible Verifier Module."""

from __future__ import annotations

import os

from typing import TYPE_CHECKING, cast

from molecule import logger, util
from molecule.api import Verifier


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from molecule.verifier.base import Schema


class Ansible(Verifier):
    """`Ansible`_ is the default test verifier.

    Molecule executes a playbook (`verify.yml`) located in the role's
    `scenario.directory`.

    ``` yaml
        verifier:
          name: ansible
    ```

    The testing can be disabled by setting ``enabled`` to False.

    ``` yaml
        verifier:
          name: ansible
          enabled: False
    ```

    Environment variables can be passed to the verifier.

    ``` yaml

        verifier:
          name: ansible
          env:
            FOO: bar
    ```
    """

    @property
    def _log(self) -> logger.ScenarioLoggerAdapter:
        """Get a fresh scenario logger with current context.

        Returns:
            A scenario logger adapter with current scenario and step context.
        """
        # Get step context from the current action being executed
        step_name = getattr(self._config, "action", "verify")
        scenario_name = self._config.scenario.name if self._config else "unknown"
        return logger.get_scenario_logger(__name__, scenario_name, step_name)

    @property
    def name(self) -> str:
        """Name of the verifier.

        Returns:
            The name of the verifier.
        """
        return "ansible"

    @property
    def default_options(self) -> MutableMapping[str, str | bool]:
        """Get default CLI arguments provided to ``cmd``.

        Returns:
            The default verifier options.
        """
        return {}

    @property
    def default_env(self) -> dict[str, str]:
        """Get default env variables provided to ``cmd``.

        Returns:
            The default verifier environment variables.
        """
        env = cast("dict[str, str]", os.environ)
        env = util.merge_dicts(env, self._config.env)
        if self._config.provisioner:
            env = util.merge_dicts(env, self._config.provisioner.env)
        return env

    def execute(self, action_args: list[str] | None = None) -> None:
        """Execute ``cmd``.

        Args:
            action_args: list of arguments to be passed.
        """
        if not self.enabled:
            msg = "Skipping, verifier is disabled."
            self._log.warning(msg)
            return

        msg = "Running Ansible Verifier"
        self._log.info(msg)

        if self._config.provisioner:
            self._config.provisioner.verify(action_args)

            msg = "Verifier completed successfully."
            self._log.info(msg)

    def schema(self) -> Schema:
        """Return validation schema.

        Returns:
            Verifier schema.
        """
        return {
            "verifier": {
                "type": "dict",
                "schema": {"name": {"type": "string", "allowed": ["ansible"]}},
            },
        }



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Status Module."""

from __future__ import annotations

from typing import NamedTuple


class Status(NamedTuple):
    """Scenario status information.

    Attributes:
        instance_name: Name of the instance.
        driver_name: Name of the driver being used.
        provisioner_name: Name of the provisioner being used.
        scenario_name: Name of the scenario being run.
        created: Has the scenario been created.
        converged: Has the scenario converged.
    """

    instance_name: str
    driver_name: str
    provisioner_name: str
    scenario_name: str
    created: str
    converged: str



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from click.testing import CliRunner

from molecule.command import converge
from molecule.shell import main


if TYPE_CHECKING:
    from unittest.mock import Mock

    import pytest

    from pytest_mock import MockerFixture

    from molecule import config


# NOTE(retr0h): The use of the `patched_config_validate` fixture, disables
# config.Config._validate from executing.  Thus preventing odd side-effects
# throughout patched.assert_called unit tests.
def test_converge_execute(  # noqa: D103
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    patched_ansible_converge: Mock,
    patched_config_validate: Any,  # noqa: ANN401
    config_instance: config.Config,
) -> None:
    c = converge.Converge(config_instance)
    c.execute()

    assert "default" in caplog.text
    assert "converge" in caplog.text

    patched_ansible_converge.assert_called_once_with()

    assert config_instance.state.converged


def test_ansible_args_passed_to_scenarios_get_configs(mocker: MockerFixture) -> None:  # noqa: D103
    # Scenarios patch is needed to safely invoke CliRunner
    # in the test environment and block scenario execution
    mocker.patch("molecule.scenarios.Scenarios")
    patched_get_configs = mocker.patch("molecule.command.base.get_configs")

    runner = CliRunner()
    args = ("converge", "--", "-e", "testvar=testvalue")
    ansible_args = args[2:]
    runner.invoke(main, args, obj={})

    # call index [0][2] is the 3rd positional argument to get_configs,
    # which should be the tuple of parsed ansible_args from the CLI
    assert patched_get_configs.call_args[0][2] == ansible_args



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Base Driver Module."""

from __future__ import annotations

import inspect
import os

from abc import ABC, abstractmethod
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING

from molecule.status import Status


if TYPE_CHECKING:
    from typing import Any

    from molecule.config import Config
    from molecule.types import DriverOptions


class Driver(ABC):
    """Driver Class.

    Attributes:
        title: Short description of the driver.
    """

    title = ""

    def __init__(self, config: Config) -> None:
        """Initialize code for all :ref:`Driver` classes.

        Args:
            config: An instance of a Molecule config.
        """
        self._config = config
        self._path = os.path.abspath(  # noqa: PTH100
            os.path.dirname(inspect.getfile(self.__class__)),  # noqa: PTH120
        )
        self.module = self.__module__.split(".", maxsplit=1)[0]
        self.version = version(self.module)

    @property
    @abstractmethod
    def name(self) -> str:  # pragma: no cover
        """Name of the driver.

        Returns:
            Name of the driver.
        """

    @name.setter
    @abstractmethod
    def name(self, value: str) -> None:  # pragma: no cover
        """Driver name setter.

        Args:
            value: New driver name.
        """

    @property
    def testinfra_options(self) -> dict[str, str]:
        """Testinfra specific options and returns a dict.

        :returns: dict
        """
        return {
            "connection": "ansible",
            "ansible-inventory": self._config.provisioner.inventory_directory,  # type: ignore[union-attr]
        }

    @property
    @abstractmethod
    def login_cmd_template(self) -> str:  # pragma: no cover
        """Get the login command template to be populated by ``login_options`` as a string.

        Returns:
            str
        """

    @property
    @abstractmethod
    def default_ssh_connection_options(self) -> list[str]:  # pragma: no cover
        """SSH client options and returns a list.

        :returns: list
        """

    @property
    @abstractmethod
    def default_safe_files(self) -> list[str]:  # pragma: no cover
        """Generate files to be preserved.

        :returns: list
        """

    @abstractmethod
    def login_options(
        self,
        instance_name: str,
    ) -> dict[str, str]:  # pragma: no cover
        """Options used in the login command and returns a dict.

        Args:
            instance_name: A string containing the instance to login to.

        Returns:
            dict
        """

    @abstractmethod
    def ansible_connection_options(
        self,
        instance_name: str,
    ) -> dict[str, Any]:  # pragma: no cover
        """Ansible specific connection options supplied to inventory and returns a dict.

        Args:
            instance_name: A string containing the instance to login to.

        Returns:
            dict
        """

    @abstractmethod
    def sanity_checks(self) -> None:
        """Confirm that driver is usable.

        Sanity checks to ensure the driver can do work successfully. For
        example, when using the Docker driver, we want to know that the Docker
        daemon is running and we have the correct Docker Python dependency.
        Each driver implementation can decide what is the most stable sanity
        check for itself.
        """

    @property
    def options(self) -> DriverOptions:
        """Driver options.

        Returns:
            Dictionary of driver options.
        """
        return self._config.config["driver"]["options"]

    @property
    def instance_config(self) -> str:
        """Instance config file location.

        Returns:
            Path to instance_config.yml.
        """
        return str(
            Path(
                self._config.scenario.ephemeral_directory,
                "instance_config.yml",
            ),
        )

    @property
    def ssh_connection_options(self) -> list[str]:
        """SSH connection options.

        Returns:
            List of ssh connection options.
        """
        if self._config.config["driver"]["ssh_connection_options"]:
            return self._config.config["driver"]["ssh_connection_options"]
        return self.default_ssh_connection_options

    @property
    def safe_files(self) -> list[str]:
        """Safe files.

        Returns:
            List of safe files.
        """
        return self.default_safe_files + self._config.config["driver"]["safe_files"]

    @property
    def delegated(self) -> bool:
        """Is the driver delegated.

        Returns:
            Whether the driver is delegated.
        """
        return self.name == "default"

    @property
    def managed(self) -> bool:
        """Is the driver is managed.

        Returns:
            Whether the driver is managed.
        """
        return self.options["managed"]

    def status(self) -> list[Status]:
        """Collect the instances state and returns a list.

        !!! note

            Molecule assumes all instances were created successfully by
            Ansible, otherwise Ansible would return an error on create.  This
            may prove to be a bad assumption.  However, configuring Molecule's
            driver to match the options passed to the playbook may prove
            difficult.  Especially in cases where the user is provisioning
            instances off localhost.

        Returns:
            Status for each instance.
        """
        status_list = []
        for platform in self._config.platforms.instances:
            instance_name = platform["name"]
            driver_name = self.name
            provisioner_name = self._config.provisioner.name if self._config.provisioner else ""
            scenario_name = self._config.scenario.name

            status_list.append(
                Status(
                    instance_name=instance_name,
                    driver_name=driver_name,
                    provisioner_name=provisioner_name,
                    scenario_name=scenario_name,
                    created=self._created(),
                    converged=self._converged(),
                ),
            )

        return status_list

    def _get_ssh_connection_options(self) -> list[str]:
        # LogLevel=ERROR is needed in order to avoid warnings like:
        # Warning: Permanently added ... to the list of known hosts.
        return [
            "-o UserKnownHostsFile=/dev/null",
            "-o ControlMaster=auto",
            "-o ControlPersist=60s",
            "-o ForwardX11=no",
            "-o LogLevel=ERROR",
            "-o IdentitiesOnly=yes",
            "-o StrictHostKeyChecking=no",
        ]

    def _created(self) -> str:
        return str(self._config.state.created).lower()

    def _converged(self) -> str:
        return str(self._config.state.converged).lower()

    def __eq__(self, other: object) -> bool:
        """Implement equality comparison.

        Args:
            other: object to compare against this one.

        Returns:
            Whether other matches the name of this driver.
        """
        # trick that allows us to test if a driver is loaded via:
        # if 'driver-name' in drivers()
        return str(self) == str(other)

    def __lt__(self, other: object) -> bool:
        """Implement lower than comparison.

        Args:
            other: object to compare against this one.

        Returns:
            Whether other is less than the name of this driver.
        """
        return str.__lt__(str(self), str(other))

    def __hash__(self) -> int:
        """Perform object hash.

        Returns:
            Hash of driver name.
        """
        return self.name.__hash__()

    def __str__(self) -> str:
        """Return readable string representation of object.

        Returns:
            Driver name.
        """
        return self.name

    def __repr__(self) -> str:
        """Return detailed string representation of object.

        Returns:
            Driver name.
        """
        return self.name

    def __rich__(self) -> str:
        """Return rich representation of object.

        Returns:
            Driver name.
        """
        return self.__str__()

    def get_playbook(self, step: str) -> str | None:
        """Return embedded playbook location or None.

        The default location is relative to the file implementing the driver
        class, allowing driver writers to define playbooks without having
        to override this method.

        Args:
            step: The step to look for a playbook for.

        Returns:
            The playbook file path, if any.
        """
        p = Path(
            self._path,
            "playbooks",
            step + ".yml",
        )
        if p.is_file():
            return str(p)
        return None

    def schema_file(self) -> str | None:  # pragma: no cover
        """Return schema file path.

        Returns:
            Path to schema file.
        """
        return None

    def modules_dir(self) -> str | None:
        """Return path to ansible modules included with driver.

        Returns:
            Path to modules dir if one exists.
        """
        p = Path(self._path, "modules")
        if p.is_dir():
            return str(p)
        return None

    def reset(self) -> None:
        """Release all resources owned by molecule.

        This is a destructive operation that would affect all resources managed
        by molecule, regardless the scenario name.  Molecule will use metadata
        like labels or tags to annotate resources allocated by it.
        """
        return

    @property
    def required_collections(self) -> dict[str, str]:
        """Return collections dict containing names and versions required."""
        return {}



"""Molecule Application Module."""

from __future__ import annotations

from functools import lru_cache
from pathlib import Path
from subprocess import CalledProcessError, CompletedProcess
from typing import TYPE_CHECKING

from ansible_compat.runtime import Runtime

from molecule.util import print_environment_vars


if TYPE_CHECKING:
    from pathlib import Path


class App:
    """App class that keep runtime status."""

    def __init__(self, path: Path) -> None:
        """Create a new app instance.

        Args:
            path: The path to the project.
        """
        self.runtime = Runtime(project_dir=path, isolated=False)

    def run_command(  # noqa: PLR0913
        self,
        cmd: str | list[str],
        env: dict[str, str] | None = None,
        cwd: Path | None = None,
        *,
        debug: bool = False,
        echo: bool = False,  # noqa: ARG002
        quiet: bool = False,  # noqa: ARG002
        check: bool = False,
    ) -> CompletedProcess[str]:
        """Execute the given command and returns None.

        Args:
            cmd: A list of strings containing the command to run.
            env: A dict containing the shell's environment.
            cwd: An optional Path to the working directory.
            debug: An optional bool to toggle debug output.
            echo: An optional bool to toggle command echo.
            quiet: An optional bool to toggle command output.
            check: An optional bool to toggle command error checking.

        Returns:
            A completed process object.

        Raises:
            CalledProcessError: If return code is nonzero and check is True.
        """
        if debug:
            print_environment_vars(env)

        result = self.runtime.run(
            args=cmd,
            env=env,
            cwd=cwd,
            tee=True,
            set_acp=False,
        )
        if result.returncode != 0 and check:
            raise CalledProcessError(
                returncode=result.returncode,
                cmd=result.args,
                output=result.stdout,
                stderr=result.stderr,
            )
        return result


@lru_cache
def get_app(path: Path) -> App:
    """Return the app instance.

    Args:
        path: The path to the project.

    Returns:
        App: The app instance.
    """
    return App(path)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Provisioner Base Module."""

from __future__ import annotations

import abc

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from molecule.config import Config


class Base(abc.ABC):
    """Provisioner Base Class."""

    def __init__(self, config: Config) -> None:
        """Initialize code for all :ref:`Provisioner` classes.

        Args:
            config: An instance of a Molecule config.
        """
        self._config = config

    @property
    @abc.abstractmethod
    def default_options(self) -> dict[str, str | bool]:  # pragma: no cover
        """Get default CLI arguments provided to ``cmd``.

        Returns:
            The default CLI arguments.
        """

    @property
    @abc.abstractmethod
    def default_env(self) -> dict[str, str]:  # pragma: no cover
        """Get default env variables provided to ``cmd``.

        Returns:
            The default env variables.
        """

    @property
    @abc.abstractmethod
    def name(self) -> str:  # pragma: no cover
        """Name of the provisioner.

        Returns:
            The provisioner's name.
        """



"""Constants used by molecule."""

from __future__ import annotations

import os


# Exit codes
RC_SUCCESS = 0
RC_TIMEOUT = 3
RC_SETUP_ERROR = 4  # Broken setup, like missing Ansible
RC_UNKNOWN_ERROR = 5  # Unexpected errors for which we do not have more specific codes, yet

# File and directory patterns
MOLECULE_HEADER = "# Molecule managed"
MOLECULE_GLOB = os.environ.get("MOLECULE_GLOB", "molecule/*/molecule.yml")

# Default values
MOLECULE_DEFAULT_SCENARIO_NAME = "default"
MOLECULE_PLATFORM_NAME = os.environ.get("MOLECULE_PLATFORM_NAME", None)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import pytest

from molecule import config, platforms


@pytest.fixture
def _instance(config_instance: config.Config):  # type: ignore[no-untyped-def]  # noqa: ANN202
    return platforms.Platforms(config_instance)


def test_instances_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = [
        {"groups": ["foo", "bar"], "name": "instance-1", "children": ["child1"]},
        {"groups": ["baz", "foo"], "name": "instance-2", "children": ["child2"]},
    ]

    assert x == _instance.instances


@pytest.fixture(name="platform_name")
def fixture_platform_name(request, config_instance: config.Config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return platforms.Platforms(config_instance, platform_name=request.param)


@pytest.mark.parametrize("platform_name", ["instance-1"], indirect=True)  # noqa: PT007
def test_instances_property_with_platform_name_instance_1(platform_name):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = [
        {"groups": ["foo", "bar"], "name": "instance-1", "children": ["child1"]},
    ]

    assert x == platform_name.instances


@pytest.mark.parametrize("platform_name", ["instance-2"], indirect=True)  # noqa: PT007
def test_instances_property_with_platform_name_instance_2(platform_name):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = [
        {"groups": ["baz", "foo"], "name": "instance-2", "children": ["child2"]},
    ]

    assert x == platform_name.instances



#  Copyright (c) 2015-2018 Cisco Systems, Inc.

#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Ansible-Playbooks Provisioner Module."""

from __future__ import annotations

import os

from pathlib import Path
from typing import TYPE_CHECKING

from molecule import logger, util


if TYPE_CHECKING:
    from typing import Literal

    from molecule.config import Config

    Section = Literal[
        "cleanup",
        "create",
        "converge",
        "destroy",
        "prepare",
        "side_effect",
        "verify",
    ]


class AnsiblePlaybooks:
    """A class to act as a module to namespace playbook properties."""

    def __init__(self, config: Config) -> None:
        """Initialize a new namespace class.

        Args:
            config: An instance of a Molecule config.
        """
        self._config = config

    @property
    def _log(self) -> logger.ScenarioLoggerAdapter:
        """Get a fresh scenario logger with current context.

        Returns:
            A scenario logger adapter with current scenario and step context.
        """
        # Get step context from the current action being executed
        step_name = getattr(self._config, "action", "provisioner")
        return logger.get_scenario_logger(__name__, self._config.scenario.name, step_name)

    @property
    def cleanup(self) -> str | None:
        """Get the cleanup playbook path.

        Returns:
            Path to cleanup.yml.
        """
        return self._get_playbook("cleanup")

    @property
    def create(self) -> str | None:
        """Get the create playbook path.

        Returns:
            Path to create.yml.
        """
        return self._get_playbook("create")

    @property
    def converge(self) -> str | None:
        """Get the converge playbook path.

        Returns:
            Path to converge.yml.
        """
        return self._get_playbook("converge")

    @property
    def destroy(self) -> str | None:
        """Get the destroy playbook path.

        Returns:
            Path to destroy.yml.
        """
        return self._get_playbook("destroy")

    @property
    def prepare(self) -> str | None:
        """Get the prepare playbook path.

        Returns:
            Path to prepare.yml.
        """
        return self._get_playbook("prepare")

    @property
    def side_effect(self) -> str | None:
        """Get the side_effect playbook path.

        Returns:
            Path to side_effect.yml.
        """
        return self._get_playbook("side_effect")

    @property
    def verify(self) -> str | None:
        """Get the verify playbook path.

        Returns:
            Path to verify.yml.
        """
        return self._get_playbook("verify")

    def _get_playbook_directory(self) -> Path:
        if self._config.provisioner:
            return util.abs_path(
                Path(self._config.provisioner.directory, "playbooks"),
            )
        return Path()

    def _get_playbook(self, section: Section) -> str | None:
        """Return path to playbook or None if playbook is not needed.

        Return None when there is no playbook configured and when action is
        considered skippable.

        Args:
            section: Named section to retrieve playbook for.

        Returns:
            The playbook path, or none if one is not needed.
        """
        c = self._config.config
        driver_dict: dict[Section, str | None] | None = c["provisioner"]["playbooks"].get(  # type: ignore[assignment]
            self._config.driver.name,
        )

        playbook: str | None = c["provisioner"]["playbooks"][section]
        if driver_dict:
            try:
                playbook = driver_dict[section]
            except KeyError as exc:
                self._log.exception(exc)
        if self._config.provisioner and playbook is not None:
            playbook = self._config.provisioner.abs_path(playbook)
            if playbook:
                playbook = self._normalize_playbook(playbook)

                if os.path.exists(playbook):  # noqa: PTH110
                    return playbook

            if os.path.exists(self._get_bundled_driver_playbook(section)):  # noqa: PTH110
                return self._get_bundled_driver_playbook(section)
            if section not in [
                # these playbooks can be considered optional
                "prepare",
                "create",
                "destroy",
                "cleanup",
                "side_effect",
                "verify",
            ]:
                return playbook
        return None

    def _get_bundled_driver_playbook(self, section: Section) -> str:
        driver_path = self._config.driver.get_playbook(section)
        if driver_path:
            return driver_path

        path = Path(
            self._get_playbook_directory(),
            self._config.driver.name,
            self._config.config["provisioner"]["playbooks"][section],
        )
        if path.exists():
            return str(path)
        path = Path(
            self._config.driver._path,  # noqa: SLF001
            "playbooks",
            self._config.config["provisioner"]["playbooks"][section],
        )
        return str(path)

    def _normalize_playbook(self, playbook: str) -> str:
        """Return current filename to use for a playbook by allowing fallbacks.

        Currently used to deprecate use of playbook.yml in favour of converge.yml

        Args:
            playbook: Playbook path to alter.

        Returns:
            Normalized playbook path.
        """
        play_path = Path(playbook)
        if not playbook or play_path.is_file():
            return playbook

        pb_rename_map = {"converge.yml": "playbook.yml"}
        basename = play_path.name
        if basename in pb_rename_map:
            fb_playbook = play_path.parent / pb_rename_map[basename]
            if fb_playbook.is_file():
                self._log.warning(
                    "%s was deprecated, rename it to %s",
                    fb_playbook.name,
                    basename,
                )
                playbook = str(fb_playbook)
        return playbook



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import copy
import os

from subprocess import CompletedProcess
from typing import TYPE_CHECKING, Any

import pytest

from molecule import config, util


if TYPE_CHECKING:
    from collections.abc import Generator, MutableMapping
    from pathlib import Path
    from unittest.mock import Mock

    from pytest_mock import MockerFixture


def write_molecule_file(file: Path, data: Any) -> None:  # noqa: ANN401, D103
    util.write_file(file, util.safe_dump(data))


def os_split(s: str) -> tuple[str, ...]:  # noqa: D103
    rest, tail = os.path.split(s)
    if rest in ("", os.path.sep):
        return (tail,)
    return (*os_split(rest), tail)


@pytest.fixture(name="molecule_data")
def fixture_molecule_data() -> dict[str, Any]:
    """Provide a default molecule data dictionary."""
    return {
        "dependency": {"name": "galaxy"},
        "driver": {"name": "default", "options": {"managed": True}},
        "platforms": [
            {"name": "instance-1", "groups": ["foo", "bar"], "children": ["child1"]},
            {"name": "instance-2", "groups": ["baz", "foo"], "children": ["child2"]},
        ],
        "provisioner": {
            "name": "ansible",
            "options": {"become": True},
            "config_options": {},
        },
        "scenario": {"name": "default"},
        "verifier": {"name": "ansible"},
    }


@pytest.fixture
def config_instance(
    monkeypatch: pytest.MonkeyPatch,
    molecule_data: dict[str, Any],
    request: pytest.FixtureRequest,
    test_cache_path: Path,
) -> Generator[config.Config, None, None]:
    """Return a Config instance.

    Merge molecule_data with any fixture that requests it and write the
    resulting data to a molecule file.  Return a Config instance with the
    molecule file path.

    Because the test_cache_path is gitignored, monkeypatch the filter_ignored_scenarios
    function to return the original list of scenarios.

    Args:
        monkeypatch: The pytest fixture for patching
        molecule_data: The pytest fixture for molecule data.
        request: The pytest fixture request
        test_cache_path: The test cache path

    Yields:
        config.Config: A Config instance
    """
    mdc: MutableMapping[str, Any] = copy.deepcopy(molecule_data)
    if hasattr(request, "param"):
        mdc = util.merge_dicts(molecule_data, request.getfixturevalue(request.param))

    monkeypatch.chdir(test_cache_path)
    molecule_dir = test_cache_path / "molecule" / "default"
    molecule_dir.mkdir(parents=True, exist_ok=True)
    molecule_file = molecule_dir / "molecule.yml"
    write_molecule_file(molecule_file, mdc)

    def _filter_ignored_scenarios(scenario_paths: list[str]) -> list[str]:
        return scenario_paths

    monkeypatch.setattr("molecule.command.base.filter_ignored_scenarios", _filter_ignored_scenarios)

    _environ = dict(os.environ)
    c = config.Config(str(molecule_file))
    c.command_args = {"subcommand": "test"}
    yield c
    # restore environ which can be modified by Config()
    os.environ.clear()
    os.environ.update(_environ)


# Mocks


@pytest.fixture
def patched_print_debug(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("molecule.util.print_debug")


@pytest.fixture
def patched_logger_info(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("logging.Logger.info")


@pytest.fixture
def patched_logger_debug(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("logging.Logger.debug")


@pytest.fixture
def patched_logger_error(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("logging.Logger.error")


@pytest.fixture
def patched_run_command(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    m = mocker.patch("molecule.app.App.run_command")
    m.return_value = CompletedProcess(
        args="foo",
        returncode=0,
        stdout="patched-run-command-stdout",
        stderr="",
    )

    return m


@pytest.fixture
def patched_ansible_converge(mocker: MockerFixture) -> Mock:  # noqa: D103
    m = mocker.patch("molecule.provisioner.ansible.Ansible.converge")
    m.return_value = "patched-ansible-converge-stdout"

    return m


@pytest.fixture
def patched_add_or_update_vars(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("molecule.provisioner.ansible.Ansible._add_or_update_vars")


@pytest.fixture
def patched_ansible_galaxy(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("molecule.dependency.ansible_galaxy.AnsibleGalaxy.execute")


@pytest.fixture
def patched_default_verifier(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("molecule.verifier.ansible.Ansible.execute")


@pytest.fixture
def patched_scenario_setup(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    mocker.patch("molecule.config.Config.env")

    return mocker.patch("molecule.scenario.Scenario._setup")


@pytest.fixture
def patched_config_validate(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("molecule.config.Config._validate")



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import os
import subprocess

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import pytest

from molecule import config, util
from molecule.command import base
from molecule.exceptions import ScenarioFailureError


if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from pytest_mock import MockerFixture

    from molecule.types import CommandArgs, MoleculeArgs


FIXTURE_DIR = Path(__file__).parent.parent.parent / "fixtures" / "unit" / "test_base"


class ExtendedBase(base.Base):
    """ExtendedBase Class."""

    def execute(self, action_args: list[str] | None = None) -> None:
        """No-op the execute method.

        Args:
            action_args: Optional list of action arguments.
        """


@pytest.fixture(name="base_class")
def fixture_base_class(
    patched_config_validate: pytest.FixtureRequest,
    config_instance: config.Config,
) -> type[ExtendedBase]:
    """Return a mocked instance of ExtendedBase.

    The use of the `patched_config_validate` fixture, disables
    config.Config._validate from executing.  Thus preventing odd side-effects
    throughout patched.assert_called unit tests (retr0h)

    Args:
        patched_config_validate: Mocked config.Config._validate function.
        config_instance: Mocked config_instance fixture.

    Returns:
        ExtendedBase: An instance of ExtendedBase instantiated with the config_instance.
    """
    return ExtendedBase


@pytest.fixture(name="instance")
def fixture_instance(
    base_class: type[ExtendedBase],
    config_instance: config.Config,
) -> ExtendedBase:
    """Return a mocked instance of ExtendedBase.

    Args:
        base_class: Mocked _base_class fixture.
        config_instance: Mocked config_instance fixture.

    Returns:
        ExtendedBase: An instance of ExtendedBase instantiated with the config_instance.
    """
    return base_class(config_instance)


@pytest.fixture(name="patched_base_setup")
def fixture_patched_base_setup(mocker: MockerFixture) -> MagicMock:
    """Mock ExtendedBase setup function.

    Args:
        mocker: pytest mocker fixture.

    Returns:
        MagicMock: Mocked setup function.
    """
    return mocker.patch("tests.unit.command.test_base.ExtendedBase._setup")


@pytest.fixture(name="patched_write_config")
def fixture_patched_write_config(mocker: MockerFixture) -> MagicMock:
    """Mock molecule write_config function.

    Args:
        mocker: pytest mocker fixture.

    Returns:
        MagicMock: Mocked write_config function.
    """
    return mocker.patch("molecule.provisioner.ansible.Ansible.write_config")


@pytest.fixture(name="patched_manage_inventory")
def fixture_patched_manage_inventory(mocker: MockerFixture) -> MagicMock:
    """Mock molecule manage_inventory function.

    Args:
        mocker: pytest mocker fixture.

    Returns:
        MagicMock: Mocked manage_inventory function.
    """
    return mocker.patch("molecule.provisioner.ansible.Ansible.manage_inventory")


@pytest.fixture(name="patched_execute_subcommand")
def fixture_patched_execute_subcommand(mocker: MockerFixture) -> MagicMock:
    """Mock molecule execute_subcommand function.

    Args:
        mocker: pytest mocker fixture.

    Returns:
        MagicMock: Mocked execute_subcommand function.
    """
    return mocker.patch("molecule.command.base.execute_subcommand")


@pytest.fixture(name="patched_execute_scenario")
def fixture_patched_execute_scenario(mocker: MockerFixture) -> MagicMock:
    """Mock molecule execute_scenario function.

    Args:
        mocker: pytest mocker fixture.

    Returns:
        MagicMock: Mocked execute_scenario function.
    """
    return mocker.patch("molecule.command.base.execute_scenario")


@pytest.fixture(name="patched_prune")
def fixture_patched_prune(mocker: MockerFixture) -> MagicMock:
    """Mock molecule prune function.

    Args:
        mocker: pytest mocker fixture.

    Returns:
        MagicMock: Mocked prune function.
    """
    return mocker.patch("molecule.scenario.Scenario.prune")


@pytest.fixture(name="patched_sysexit")
def fixture_patched_sysexit(mocker: MockerFixture) -> MagicMock:
    """Mock molecule util.sysexit function.

    Args:
        mocker: pytest mocker fixture.

    Returns:
        MagicMock: Mocked util.sysexit function.
    """
    return mocker.patch("molecule.util.sysexit")


def test_command_config_private_member(instance: ExtendedBase) -> None:
    """Ensure the _config private member is a config.Config instance.

    Args:
        instance: An instance of ExtendedBase.
    """
    assert isinstance(instance._config, config.Config)


def test_init_calls_setup(
    patched_base_setup: MagicMock,
    instance: ExtendedBase,
) -> None:
    """Ensure the setup method is called during initialization.

    Args:
        patched_base_setup: Mocked setup function.
        instance: An instance of ExtendedBase.
    """
    assert isinstance(instance, ExtendedBase)
    patched_base_setup.assert_called_once_with()


def test_command_setup(
    patched_write_config: MagicMock,
    patched_manage_inventory: MagicMock,
    instance: ExtendedBase,
) -> None:
    """Ensure the setup method runs normally.

    Args:
        patched_write_config: Mocked write_config function.
        patched_manage_inventory: Mocked manage_inventory function.
        instance: An instance of ExtendedBase.

    """
    assert Path(instance._config.provisioner.inventory_file).parent.is_dir()  # type: ignore[union-attr]
    assert Path(instance._config.config_file).is_file()

    patched_manage_inventory.assert_called_once_with()
    patched_write_config.assert_called_once_with()


@pytest.mark.usefixtures("config_instance")
def test_execute_cmdline_scenarios(patched_execute_scenario: MagicMock) -> None:
    """Ensure execute_cmdline_scenarios runs normally.

    Execute_scenario is called once, indicating the function correctly loops over Scenarios.

    Args:
        patched_execute_scenario: Mocked execute_scenario function.
    """
    scenario_name = None
    args: MoleculeArgs = {}
    command_args: CommandArgs = {"destroy": "always", "subcommand": "test"}
    base.execute_cmdline_scenarios(scenario_name, args, command_args)

    scenario_count = 1
    assert patched_execute_scenario.call_count == scenario_count


@pytest.mark.usefixtures("config_instance")
def test_execute_cmdline_scenarios_missing(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Confirm execute_scenario exits properly when given a nonexistent scenario.

    Args:
        caplog: pytest caplog fixture.
    """
    scenario_name = ["nonexistent"]
    args: MoleculeArgs = {}
    command_args: CommandArgs = {"destroy": "always", "subcommand": "test"}

    with pytest.raises(SystemExit):
        base.execute_cmdline_scenarios(scenario_name, args, command_args)

    error_msg = "'molecule/nonexistent/molecule.yml' glob failed.  Exiting."
    assert error_msg in caplog.text


@pytest.mark.usefixtures("config_instance")
def test_execute_cmdline_scenarios_prune(
    patched_execute_subcommand: MagicMock,
    patched_prune: MagicMock,
) -> None:
    """Confirm prune is called when destroy is 'always'.

    Args:
        patched_execute_subcommand: Mocked execute_subcommand function.
        patched_prune: Mocked prune function.
    """
    scenario_name = ["default"]
    args: MoleculeArgs = {}
    command_args: CommandArgs = {"destroy": "always", "subcommand": "test"}

    base.execute_cmdline_scenarios(scenario_name, args, command_args)

    assert patched_execute_subcommand.called
    assert patched_prune.called


@pytest.mark.usefixtures("config_instance")
def test_execute_cmdline_scenarios_no_prune(
    patched_prune: MagicMock,
    patched_execute_subcommand: MagicMock,
) -> None:
    """Confirm prune is not called when destroy is 'never'.

    Args:
        patched_prune: Mocked prune function.
        patched_execute_subcommand: Mocked execute_subcommand function.
    """
    scenario_name = ["default"]
    args: MoleculeArgs = {}
    command_args: CommandArgs = {"destroy": "never", "subcommand": "test"}

    base.execute_cmdline_scenarios(scenario_name, args, command_args)

    assert patched_execute_subcommand.called
    assert not patched_prune.called


@pytest.mark.parametrize(
    ("destroy", "subcommands"),
    (
        ("always", ("cleanup", "destroy")),
        ("never", ()),
    ),
)
@pytest.mark.usefixtures("config_instance")
def test_execute_cmdline_scenarios_exit_destroy(  # noqa: PLR0913
    patched_execute_scenario: MagicMock,
    patched_prune: MagicMock,
    patched_execute_subcommand: MagicMock,
    patched_sysexit: MagicMock,
    destroy: Literal["always", "never"],
    subcommands: tuple[str, ...],
) -> None:
    """Ensure execute_cmdline_scenarios handles errors correctly when 'destroy' is set.

    - destroy subcommand is only run when destroy == "always"
    - scenario is only pruned when destroy == "always"

    Args:
        patched_execute_scenario: Mocked execute_scenario function.
        patched_prune: Mocked prune function.
        patched_execute_subcommand: Mocked execute_subcommand function.
        patched_sysexit: Mocked util.sysexit function.
        destroy: Value to set 'destroy' arg to.
        subcommands: Expected subcommands to run after execute_scenario fails.
    """
    scenario_name = ["default"]
    args: MoleculeArgs = {}
    command_args: CommandArgs = {"destroy": destroy, "subcommand": "test"}
    patched_execute_scenario.side_effect = ScenarioFailureError()

    base.execute_cmdline_scenarios(scenario_name, args, command_args)
    assert patched_execute_scenario.called

    call_count = len(subcommands)
    assert patched_execute_subcommand.call_count == call_count
    for i, subcommand in enumerate(subcommands):
        assert patched_execute_subcommand.call_args_list[i][0][1] == subcommand
    assert patched_prune.called is (destroy == "always")
    assert patched_sysexit.called


def test_execute_subcommand(config_instance: config.Config) -> None:
    """Ensure execute_subcommand runs normally.

    Scenario's config.action is mutated in-place for every sequence action,
        so make sure that is currently set to the executed action

    Args:
        config_instance: Mocked config_instance fixture.
    """
    assert config_instance.action != "list"
    assert base.execute_subcommand(config_instance, "list")
    assert config_instance.action == "list"


def test_execute_scenario(
    mocker: MockerFixture,
    patched_execute_subcommand: MagicMock,
) -> None:
    """Ensure execute_scenario runs normally.

    - call a spoofed scenario with a sequence that does not include destroy
    - execute_subcommand should be called once for each sequence item
    - prune should not be called, since the sequence has no destroy step

    Args:
        mocker: pytest mocker fixture.
        patched_execute_subcommand: Mocked execute_subcommand function.
    """
    scenario = mocker.Mock()
    scenario.sequence = ("a", "b", "c")

    base.execute_scenario(scenario)

    assert patched_execute_subcommand.call_count == len(scenario.sequence)
    assert not scenario.prune.called


def test_execute_scenario_destroy(
    mocker: MockerFixture,
    patched_execute_subcommand: MagicMock,
) -> None:
    """Ensure execute_scenario runs normally.

    - call a spoofed scenario with a sequence that includes destroy
    - execute_subcommand should be called once for each sequence item
    - prune should be called, since the sequence has a destroy step

    Args:
        mocker: pytest mocker fixture.
        patched_execute_subcommand: Mocked execute_subcommand function.
    """
    scenario = mocker.Mock()
    scenario.sequence = ("a", "b", "destroy", "c")

    base.execute_scenario(scenario)

    assert patched_execute_subcommand.call_count == len(scenario.sequence)
    assert scenario.prune.called


def test_execute_scenario_shared_destroy(
    mocker: MockerFixture,
    patched_execute_subcommand: MagicMock,
) -> None:
    """Ensure execute_scenario runs normally.

    - call a spoofed scenario with a sequence that includes destroy
    - shared_state is set which means destroy should be ignored.
    - execute_subcommand should be called once for each sequence item
    - prune should not be called, since destroy has been elided from the sequence.

    Args:
        mocker: pytest mocker fixture.
        patched_execute_subcommand: Mocked execute_subcommand function.
    """
    scenario = mocker.Mock()
    scenario.config.shared_data = True
    scenario.sequence = ("a", "b", "destroy", "c")
    expected_sequence = ("a", "b", "c")

    base.execute_scenario(scenario)

    assert patched_execute_subcommand.call_count == len(expected_sequence)
    assert not scenario.prune.called


def test_get_configs(config_instance: config.Config) -> None:
    """Ensure get_configs returns a list of config.Config instances.

    Args:
        config_instance: Mocked config_instance fixture.
    """
    molecule_file = config_instance.molecule_file
    data = config_instance.config
    util.write_file(molecule_file, util.safe_dump(data))

    result = base.get_configs({}, {})
    assert len(result) == 1
    assert isinstance(result, list)
    assert isinstance(result[0], config.Config)


def test_verify_configs(config_instance: config.Config) -> None:
    """Ensure verify_configs runs normally and does not raise.

    Args:
        config_instance: Mocked config_instance fixture.
    """
    configs = [config_instance]
    base._verify_configs(configs)


def test_verify_configs_raises_with_no_configs(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Ensure verify_configs raises when no configs are provided.

    Args:
        caplog: pytest caplog fixture.
    """
    with pytest.raises(ScenarioFailureError) as e:
        base._verify_configs([])

    assert e.value.code == 1

    msg = "'molecule/*/molecule.yml' glob failed.  Exiting."
    assert msg in caplog.text


def test_verify_configs_raises_with_duplicate_configs(
    caplog: pytest.LogCaptureFixture,
    config_instance: config.Config,
) -> None:
    """Ensure verify_configs raises when duplicate configs are provided.

    Args:
        caplog: pytest caplog fixture.
        config_instance: Mocked config_instance fixture.
    """
    configs = [config_instance, config_instance]

    with pytest.raises(ScenarioFailureError) as e:
        base._verify_configs(configs)

    assert e.value.code == 1

    msg = "Duplicate scenario name 'default' found.  Exiting."
    assert msg in caplog.text


def test_get_subcommand() -> None:
    """Ensure get_subcommand returns the subcommand name."""
    assert base._get_subcommand(__name__) == "test_base"


@pytest.mark.parametrize(
    "shell",
    [  # noqa: PT007
        "bash",
        "zsh",
        "fish",
    ],
)
def test_command_completion(shell: str) -> None:
    """Ensure shell completion runs normally.

    Args:
        shell: Shell name.
    """
    env = os.environ.copy()
    env["_MOLECULE_COMPLETE"] = f"{shell}_source"
    bash_version = "0.0"

    if "bash" in shell:
        bash_version = subprocess.run(
            ["bash", "--version"],
            text=True,
            encoding="utf-8",
            capture_output=True,
            check=False,
        ).stdout.split()[3][0:3]

    result = subprocess.run(
        ["molecule"],
        env=env,
        text=True,
        capture_output=True,
        check=False,
    )

    if "bash" in shell and (float(bash_version) < 4.4):  # noqa: PLR2004
        string = "Shell completion is not supported for Bash versions older than 4.4"
        assert string not in result.stdout
    else:
        assert result.returncode == 0
        assert "Found config file" not in result.stdout



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Logging Module."""

from __future__ import annotations

import logging
import os
import time

from functools import wraps
from typing import TYPE_CHECKING, Protocol, cast

from ansible_compat.ports import cache

from molecule.ansi_output import AnsiOutput
from molecule.console import console, original_stderr
from molecule.text import underscore


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, MutableMapping
    from typing import ParamSpec, TypeVar

    from molecule.config import Config

    P = ParamSpec("P")
    R = TypeVar("R")


LOG_LEVEL_LUT = {
    0: logging.INFO,
    1: logging.DEBUG,
}


class HasConfig(Protocol):
    """A class with a _config attribute.

    There are a few such classes in Molecule. We just care that it's one of them.

    Attributes:
        _config: A Config instance.
    """

    _config: Config


class MoleculeConsoleHandler(logging.Handler):
    """Custom logging handler that uses ANSI color codes directly.

    Provides colored output without depending on RichHandler or competing formatters.
    """

    def __init__(
        self,
        *,
        show_time: bool = False,
        show_path: bool = False,
    ) -> None:
        """Initialize the console handler.

        Args:
            show_time: Whether to show timestamps.
            show_path: Whether to show file paths.
        """
        super().__init__()
        self.show_time = show_time
        self.show_path = show_path
        self.ansi_output = AnsiOutput()

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log record with scenario context using original stderr.

        All messages go to the original stderr, bypassing Rich's redirection entirely.
        Also updates the log record with a plain text version for caplog compatibility.

        Args:
            record: The logging record to emit.
        """
        try:
            # Format the message
            message = self.format(record)

            # Check if this message has scenario context passed from ScenarioLoggerAdapter
            scenario_name = getattr(record, "molecule_scenario", None)
            step_name = getattr(record, "molecule_step", None)

            # Get both colored and plain versions using dual methods
            # cspell:ignore levelname
            colored_level, plain_level = self.ansi_output.format_log_level(record.levelname)

            # Process message once for each version
            colored_message = self.ansi_output.process_markup(message)
            plain_message = self.ansi_output.strip_markup(message)

            if scenario_name:
                colored_scenario, plain_scenario = self.ansi_output.format_scenario(
                    scenario_name,
                    step_name,
                )

                # Create colored output for display
                colored_output = f"{colored_level} {colored_scenario} {colored_message}"

                # Create plain output for caplog
                plain_output = f"{plain_level} {plain_scenario} {plain_message}"
            else:
                # Create colored output for display
                colored_output = f"{colored_level} {colored_message}"

                # Create plain output for caplog
                plain_output = f"{plain_level} {plain_message}"

            # Update the log record with the plain version for caplog and other handlers
            record.msg = plain_output
            record.args = ()

            # Write colored output to stderr for users
            original_stderr.write(colored_output + "\n")
            original_stderr.flush()

        except Exception:  # noqa: BLE001
            self.handleError(record)

    def format(self, record: logging.LogRecord) -> str:
        """Format the log record message.

        Args:
            record: The logging record to format.

        Returns:
            The formatted message string.
        """
        # Use the default formatter if one is set
        if self.formatter:
            return self.formatter.format(record)

        # Simple default formatting
        return record.getMessage()


def configure() -> None:
    """Configure a molecule root logger.

    All other loggers will inherit the configuration we set here.
    """
    # Keep using root logger because we do want to process messages from other
    # libraries.
    logger = logging.getLogger()

    # Use our custom ANSI color handler instead of RichHandler
    handler = MoleculeConsoleHandler(
        show_time=False,
        show_path=False,
    )

    # Set a minimal formatter since we handle styling in the handler
    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    logger.propagate = False
    logger.setLevel(logging.INFO)


def set_log_level(log_level: int, debug: bool) -> None:  # noqa: FBT001
    """Set logging level.

    Args:
        log_level: verbosity control (0 - INFO, 1 - DEBUG)
        debug: debug mode indicator
    """
    # If we get verbosity level > 1, we just use debug because this is the
    # most detailed log level we have.
    if debug:
        log_level = 1  # DEBUG from the LOG_LEVEL_LUT
    logging.getLogger("molecule").setLevel(LOG_LEVEL_LUT.get(log_level, logging.DEBUG))


def get_logger(name: str) -> logging.Logger:
    """Return a child logger.

    Returned logger inherits configuration from the molecule logger.

    Args:
        name: Name of the child logger.

    Returns:
        A child logger of molecule.
    """
    return logging.getLogger("molecule." + name)


class ScenarioLoggerAdapter(logging.LoggerAdapter):  # type: ignore[type-arg]
    """Logger adapter that automatically includes scenario context in messages.

    This adapter includes scenario names and optional step information in log messages
    to provide better context when multiple scenarios are running or being processed.
    """

    def process(
        self,
        msg: object,
        kwargs: MutableMapping[str, object],
    ) -> tuple[object, MutableMapping[str, object]]:
        """Process the logging record.

        Args:
            msg: The log message.
            kwargs: Additional keyword arguments.

        Returns:
            A tuple of (processed_message, kwargs).
        """
        scenario_name = self.extra.get("scenario_name", "unknown") if self.extra else "unknown"
        step_name = self.extra.get("step_name") if self.extra else None

        # Pass scenario and step information through kwargs
        # Create new extra dict or copy existing one to avoid modifying the original
        current_extra = kwargs.get("extra", {})
        new_extra = dict(current_extra) if isinstance(current_extra, dict) else {}
        new_extra["molecule_scenario"] = scenario_name
        if step_name:
            new_extra["molecule_step"] = step_name
        kwargs["extra"] = new_extra
        return msg, kwargs


def get_scenario_logger(
    name: str,
    scenario_name: str,
    step_name: str,
) -> ScenarioLoggerAdapter:
    """Return a scenario-aware logger that includes scenario name in all messages.

    Args:
        name: Name of the child logger.
        scenario_name: Name of the scenario for context.
        step_name: Step name (e.g., 'converge', 'create', 'destroy').

    Returns:
        A ScenarioLoggerAdapter that includes scenario context in all messages.
    """
    logger = get_logger(name)
    extra = {"scenario_name": scenario_name, "step_name": step_name}
    return ScenarioLoggerAdapter(logger, extra)


def github_actions_groups(func: Callable[P, R]) -> Callable[P, R]:
    """Print group indicators before/after execution of a method.

    Args:
        func: The function to wrap.

    Returns:
        The wrapped function.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        self = cast("HasConfig", args[0])
        scenario = self._config.scenario.name
        subcommand = underscore(self.__class__.__name__)
        console.print(
            "::group::",
            f"[ci_info]Molecule[/] [scenario]{scenario}[/] > [action]{subcommand}[/]",
            sep="",
            markup=True,
            emoji=False,
            highlight=False,
        )
        try:
            return func(*args, **kwargs)
        finally:
            console.print("::endgroup::", markup=True, emoji=False, highlight=False)

    return wrapper


def gitlab_ci_sections(func: Callable[P, R]) -> Callable[P, R]:
    """Print group indicators before/after execution of a method.

    Args:
        func: The function to wrap.

    Returns:
        The wrapped function.
    """
    # GitLab requires:
    #  - \r (carriage return)
    #  - \e[0K (clear line ANSI escape code. We use \033 for the \e escape char)
    clear_line = "\r\033[0K"

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        self = cast("HasConfig", args[0])
        scenario = self._config.scenario.name
        subcommand = underscore(self.__class__.__name__)
        console.print(
            f"section_start:{int(time.time())}:{scenario}.{subcommand}",
            end=clear_line,
            markup=False,
            emoji=False,
            highlight=False,
        )
        console.print(
            # must be one color for the whole line or gitlab sets odd widths to each word.
            f"[ci_info]Molecule {scenario} > {subcommand}[/]",
            end="\n",
            markup=True,
            emoji=False,
            highlight=False,
        )
        try:
            return func(*args, **kwargs)
        finally:
            console.print(
                f"section_end:{int(time.time())}:{scenario}.{subcommand}",
                end=f"{clear_line}\n",
                markup=False,
                emoji=False,
                highlight=False,
            )

    return wrapper


def travis_ci_folds(func: Callable[P, R]) -> Callable[P, R]:
    """Print group indicators before/after execution of a method.

    Args:
        func: The function to wrap.

    Returns:
        The wrapped function.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        self = cast("HasConfig", args[0])
        scenario = self._config.scenario.name
        subcommand = underscore(self.__class__.__name__)
        console.print(
            f"travis_fold:start:{scenario}.{subcommand}",
            f"[ci_info]Molecule[/] [scenario]{scenario}[/] > [action]{subcommand}[/]",
            sep="",
            markup=True,
            emoji=False,
            highlight=False,
        )
        try:
            return func(*args, **kwargs)
        finally:
            console.print(
                f"travis_fold:end:{scenario}.{subcommand}",
                markup=False,
                emoji=False,
                highlight=False,
            )

    return wrapper


def section_logger(func: Callable[P, R]) -> Callable[P, R]:
    """Wrap effective execution of a method.

    Args:
        func: The function to wrap.

    Returns:
        The wrapped function.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        self = cast("HasConfig", args[0])
        step_name = underscore(self.__class__.__name__)
        scenario_log = get_scenario_logger(__name__, self._config.scenario.name, step_name)
        scenario_log.info("[exec_phase]Starting[/]")
        rt = func(*args, **kwargs)
        scenario_log.info("[exec_phase]Completed[/]")
        return rt

    return wrapper


@cache
def get_section_loggers() -> Iterable[Callable[[Callable[..., object]], Callable[..., object]]]:
    """Return a list of section wrappers to be added.

    Returns:
        A list of logging decorators.
    """
    default_section_loggers = [section_logger]
    if not os.getenv("CI"):
        return default_section_loggers
    if os.getenv("GITHUB_ACTIONS"):
        return [github_actions_groups, *default_section_loggers]
    if os.getenv("GITLAB_CI"):
        return [gitlab_ci_sections, *default_section_loggers]
    if os.getenv("TRAVIS"):
        return [travis_ci_folds, *default_section_loggers]
    # CI is set but no extra section_loggers apply.
    return default_section_loggers



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import logging

from typing import TYPE_CHECKING

import pytest

from molecule.command import check


if TYPE_CHECKING:
    from unittest.mock import MagicMock, Mock

    from pytest_mock import MockerFixture

    from molecule import config


@pytest.fixture
def _patched_ansible_check(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("molecule.provisioner.ansible.Ansible.check")


# NOTE(retr0h): The use of the `patched_config_validate` fixture, disables
# config.Config._validate from executing.  Thus preventing odd side-effects
# throughout patched.assert_called unit tests.
def test_check_execute(  # noqa: D103
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    _patched_ansible_check: Mock,  # noqa: PT019
    patched_config_validate: Mock,
    config_instance: config.Config,
) -> None:
    # Configure caplog to capture the correct logger
    with caplog.at_level(logging.INFO, logger="molecule.molecule.logger"):
        c = check.Check(config_instance)
        c.execute()

        # Check that logs contain scenario and step information in extras
        assert len(caplog.records) >= 1
        record = caplog.records[0]
        assert hasattr(record, "molecule_scenario")
        assert record.molecule_scenario == "default"
        assert hasattr(record, "molecule_step")
        assert record.molecule_step == "check"

    _patched_ansible_check.assert_called_once_with()



#  Copyright (c) 2019 Red Hat, Inc.
#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Molecule version information."""

from __future__ import annotations


try:
    from ._version import (  # type: ignore[import-not-found, unused-ignore]
        __version__,
        version,
    )
except ImportError:  # pragma: no cover
    __version__ = "0.1.dev1"
    version = __version__

__all__ = ("__version__", "version")



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Base class used by init scenario command."""

from __future__ import annotations

import json
import os
import sys

from pathlib import Path
from typing import TYPE_CHECKING

import click

from molecule import api, logger
from molecule.click_cfg import click_command_ex
from molecule.command.init import base
from molecule.config import (
    DEFAULT_DRIVER,
    MOLECULE_EMBEDDED_DATA_DIR,
    Config,
    molecule_directory,
)
from molecule.constants import MOLECULE_DEFAULT_SCENARIO_NAME
from molecule.exceptions import MoleculeError


if TYPE_CHECKING:
    from typing import TypedDict

    class CommandArgs(TypedDict):
        """Argument dictionary to pass to init-scenario playbook.

        Attributes:
            dependency_name: Name of the dependency to initialize.
            driver_name: Name of the driver to initialize.
            provisioner_name: Name of the provisioner to initialize.
            scenario_name: Name of the scenario to initialize.
            subcommand: Name of subcommand to initialize.
        """

        dependency_name: str
        driver_name: str
        provisioner_name: str
        scenario_name: str
        subcommand: str


class Scenario(base.Base):
    """Scenario Class.

    .. program:: molecule init scenario bar --role-name foo

    .. option:: molecule init scenario bar --role-name foo

        Initialize a new scenario. In order to customize the role, please refer
        to the `init role` command.

    .. program:: cd foo; molecule init scenario bar --role-name foo

    .. option:: cd foo; molecule init scenario bar --role-name foo

        Initialize an existing role with Molecule:

    .. program:: cd foo; molecule init scenario bar --role-name foo

    .. option:: cd foo; molecule init scenario bar --role-name foo

        Initialize a new scenario using a embedded template.
    """

    def __init__(self, config: Config, command_args: CommandArgs) -> None:
        """Construct Scenario.

        Args:
            config: An instance of a Molecule config.
            command_args: Arguments to pass to init-scenario playbook.
        """
        self._config = config
        self._command_args = command_args
        # For init scenario, use the scenario name from command args
        scenario_name = command_args.get("scenario_name", "unknown")
        self._log = logger.get_scenario_logger(__name__, scenario_name, "init")

    def execute(self, action_args: list[str] | None = None) -> None:  # noqa: ARG002
        """Execute the actions necessary to perform a `molecule init scenario`.

        Args:
            action_args: Arguments for this command. Unused.

        Raises:
            MoleculeError: when the scenario cannot be created.
        """
        scenario_name = self._command_args["scenario_name"]

        msg = f"Initializing new scenario {scenario_name}..."
        self._log.info(msg)
        molecule_path = Path(molecule_directory(Path.cwd()))
        scenario_directory = molecule_path / scenario_name

        if scenario_directory.is_dir():
            msg = f"The directory molecule/{scenario_name} exists. Cannot create new scenario."
            raise MoleculeError(msg)

        extra_vars = json.dumps(self._command_args)
        cmd = [
            "ansible-playbook",
            "-i",
            "localhost,",
            "--connection=local",
            "--extra-vars",
            extra_vars,
            f"{MOLECULE_EMBEDDED_DATA_DIR}/init-scenario.yml",
        ]
        env = os.environ.copy()
        # As ansible fails to find a terminal when run by molecule, we force
        # it to use colors.
        env["ANSIBLE_FORCE_COLOR"] = "1"
        env["ANSIBLE_PYTHON_INTERPRETER"] = sys.executable
        self._config.app.run_command(cmd, env=env, check=True)

        msg = f"Initialized scenario in {scenario_directory} successfully."
        self._log.info(msg)


def _role_exists(
    ctx: click.Context,  # noqa: ARG001
    param: str | None,  # noqa: ARG001
    value: str,
) -> str:  # pragma: no cover
    # if role name was not mentioned we assume that current directory is the
    # one hosting the role and determining the role name.
    if not value:
        value = Path.cwd().name

    role_directory = Path.cwd().parent / value
    if not role_directory.exists():
        msg = f"The role '{value}' not found. Please choose the proper role name."
        raise MoleculeError(msg)
    return value


@click_command_ex()
@click.pass_context
@click.option(
    "--dependency-name",
    type=click.Choice(["galaxy"]),
    default="galaxy",
    help="Name of dependency to initialize. (galaxy)",
)
@click.option(
    "--driver-name",
    "-d",
    type=str,
    default=DEFAULT_DRIVER,
    help=f"Name of driver to initialize. ({DEFAULT_DRIVER})",
)
@click.option(
    "--provisioner-name",
    type=click.Choice(["ansible"]),
    default="ansible",
    help="Name of provisioner to initialize. (ansible)",
)
@click.argument(
    "scenario-name",
    default=MOLECULE_DEFAULT_SCENARIO_NAME,
    required=False,
)
def scenario(
    ctx: click.Context,  # noqa: ARG001
    dependency_name: str,
    driver_name: str,
    provisioner_name: str,
    scenario_name: str,
) -> None:  # pragma: no cover
    """Initialize a new scenario for use with Molecule.

    If name is not specified the 'default' value will be used.

    Args:
        ctx: Click context object holding commandline arguments.
        dependency_name: Name of dependency to initialize.
        driver_name: Name of driver to use.
        provisioner_name: Name of provisioner to use.
        scenario_name: Name of scenario to initialize.

    Raises:
        click.Abort: If the specified driver is not available.
    """
    config = Config("", args={})
    available_drivers = list(api.drivers(config).keys())
    if driver_name not in available_drivers:
        if len(available_drivers) == 1 and available_drivers[0] == "default":
            click.echo(
                click.style(
                    f"Driver '{driver_name}' not available.\n\n"
                    f"Install cloud drivers with:\n"
                    f"  pip install molecule-plugins\n\n"
                    f"Currently available drivers: {available_drivers}\n",
                    fg="red",
                ),
                err=True,
            )
        else:
            click.echo(
                click.style(
                    f"Driver '{driver_name}' not available.\n"
                    f"Available drivers: {available_drivers}",
                    fg="red",
                ),
                err=True,
            )
        raise click.Abort

    command_args: CommandArgs = {
        "dependency_name": dependency_name,
        "driver_name": driver_name,
        "provisioner_name": provisioner_name,
        "scenario_name": scenario_name,
        "subcommand": __name__,
    }

    config = Config(
        "",
        args={},
    )
    s = Scenario(config, command_args)
    s.execute()



"""Console and terminal utilities."""

from __future__ import annotations

import os
import sys

from typing import Any

from enrich.console import Console
from rich.style import Style
from rich.theme import Theme

from molecule.ansi_output import should_do_markup


theme = Theme(
    {
        "info": "dim cyan",
        "warning": "magenta",
        "danger": "bold red",
        "scenario": "green",
        "action": "green",
        "section_title": "bold cyan",
        "logging.level.notset": Style(dim=True),
        "logging.level.debug": Style(color="white", dim=True),
        "logging.level.info": Style(color="blue"),
        "logging.level.warning": Style(color="red"),
        "logging.level.error": Style(color="red", bold=True),
        "logging.level.critical": Style(color="red", bold=True),
        "logging.level.success": Style(color="green", bold=True),
    },
)


# Define ANSIBLE_FORCE_COLOR if markup is enabled and another value is not
# already given. This assures that Ansible subprocesses are still colored,
# even if they do not run with a real TTY.
if should_do_markup():
    os.environ["ANSIBLE_FORCE_COLOR"] = os.environ.get("ANSIBLE_FORCE_COLOR", "1")

# Capture original streams before Rich Console objects hijack them
# These will be used by the logger to write directly to the original streams
original_stdout = sys.stdout
original_stderr = sys.stderr

console_options: dict[str, Any] = {"emoji": False, "theme": theme, "soft_wrap": True}

console = Console(
    force_terminal=should_do_markup(),
    theme=theme,
    record=True,
    redirect=True,
)
console_options_stderr = console_options.copy()
console_options_stderr["stderr"] = True
console_stderr: Console = Console(**console_options_stderr)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Delegated Driver Module."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from molecule import util
from molecule.api import Driver
from molecule.data import __file__ as data_module


if TYPE_CHECKING:
    from typing import Any

    from molecule.config import Config


class Delegated(Driver):
    r"""The class responsible for managing default instances.

    Delegated is the default driver used in Molecule.

    Under this driver, it is the developers responsibility to implement the
    create and destroy playbooks.  ``Managed`` is the default behavior of all
    drivers.

    ``` yaml
        driver:
          name: de
    ```

    However, the developer must adhere to the instance-config API. The
    developer's create playbook must provide the following instance-config
    data, and the developer's destroy playbook must reset the instance-config.

    ``` yaml

        - address: ssh_endpoint
          identity_file: ssh_identity_file  # mutually exclusive with password
          instance: instance_name
          port: ssh_port_as_string
          user: ssh_user
          shell_type: sh
          password: ssh_password  # mutually exclusive with identity_file
          become_method: valid_ansible_become_method  # optional
          become_pass: password_if_required  # optional

        - address: winrm_endpoint
          instance: instance_name
          connection: 'winrm'
          port: winrm_port_as_string
          user: winrm_user
          password: winrm_password
          shell_type: powershell
          winrm_transport: ntlm/credssp/kerberos
          winrm_cert_pem: <path to the credssp public certificate key>
          winrm_cert_key_pem: <path to the credssp private certificate key>
          winrm_server_cert_validation: validate/ignore
    ```

    This article covers how to configure and use WinRM with Ansible:
    https://docs.ansible.com/ansible/latest/user_guide/windows_winrm.html

    Molecule can also skip the provisioning/deprovisioning steps.  It is the
    developers responsibility to manage the instances, and properly configure
    Molecule to connect to said instances.

    ``` yaml
        driver:
          name: default
          options:
            managed: False
            login_cmd_template: 'docker exec -ti {instance} bash'
            ansible_connection_options:
              ansible_connection: docker
        platforms:
          - name: instance-docker
    ```

    ``` bash
        $ docker run \
            -d \
            --name instance-docker \
            --hostname instance-docker \
            -it molecule_local/ubuntu:latest sleep infinity & wait
    ```

    Use Molecule with delegated instances, which are accessible over ssh.

    !!! note

        It is the developer's responsibility to configure the ssh config file.

    ``` yaml
        driver:
          name: default
          options:
            managed: False
            login_cmd_template: 'ssh {instance} -F /tmp/ssh-config'
            ansible_connection_options:
              ansible_connection: ssh
              ansible_ssh_common_args: '-F /path/to/ssh-config'
        platforms:
          - name: instance
    ```

    Provide the files Molecule will preserve post ``destroy`` action.

    ``` yaml
        driver:
          name: default
          safe_files:
            - foo
    ```
    And in order to use localhost as molecule's target:

    ``` yaml
        driver:
          name: default
          options:
            managed: False
            ansible_connection_options:
              ansible_connection: local
    ```

    Attributes:
        title: Short description of the driver.
    """

    title = "Default driver, user is expected to manage provisioning of test resources."

    def __init__(self, config: Config) -> None:
        """Construct Delegated.

        Args:
            config: An instance of a Molecule config.
        """
        super().__init__(config)
        self._name = "default"

    @property
    def name(self) -> str:
        """Name of the driver.

        Returns:
            Name of the driver.
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Driver name setter.

        Args:
            value: New name of the driver.
        """
        self._name = value

    @property
    def login_cmd_template(self) -> str:
        """Get the login command template to be populated by ``login_options`` as a string.

        Returns:
            The login command template, if any.
        """
        if "login_cmd_template" in self.options:
            return self.options["login_cmd_template"]

        if self.managed:
            connection_options = " ".join(self.ssh_connection_options)

            return (
                f"ssh {{address}} -l {{user}} -p {{port}} -i {{identity_file}} {connection_options}"
            )
        return ""

    @property
    def default_safe_files(self) -> list[str]:
        """Generate files to be preserved.

        Returns:
            List of files to be preserved.
        """
        return []

    @property
    def default_ssh_connection_options(self) -> list[str]:
        """SSH client options.

        Returns:
            List of SSH connection options.
        """
        if self.managed:
            ssh_connopts = self._get_ssh_connection_options()
            if config_connopts := self.options.get("ansible_connection_options", {}).get(
                "ansible_ssh_common_args",
            ):
                ssh_connopts.append(
                    config_connopts,
                )
            return ssh_connopts
        return []

    def login_options(self, instance_name: str) -> dict[str, str]:
        """Login options.

        Args:
            instance_name: The name of the instance to look up login options for.

        Returns:
            Dictionary of options related to logging into the instance.
        """
        if self.managed:
            d = {"instance": instance_name}

            return util.merge_dicts(d, self._get_instance_config(instance_name))
        return {"instance": instance_name}

    def ansible_connection_options(
        self,
        instance_name: str,
    ) -> dict[str, Any]:
        """Ansible connection options.

        Args:
            instance_name: The name of the instance to look up Ansible connection options for.

        Returns:
            Dictionary of options related to ansible connection to the instance.
        """
        # list of tuples describing mappable instance params and default values
        instance_params = [
            ("become_pass", None),
            ("become_method", False),
            ("winrm_transport", None),
            ("winrm_cert_pem", None),
            ("winrm_cert_key_pem", None),
            ("winrm_server_cert_validation", None),
            ("shell_type", None),
            ("connection", "smart"),
        ]
        if self.managed:
            try:
                d = self._get_instance_config(instance_name)
                conn_dict = {}
                # Check if optional mappable params are in the instance config
                for i in instance_params:
                    if d.get(i[0], i[1]):
                        conn_dict["ansible_" + i[0]] = d.get(i[0], i[1])

                conn_dict["ansible_user"] = d.get("user")
                conn_dict["ansible_host"] = d.get("address")
                conn_dict["ansible_port"] = d.get("port")

                if d.get("identity_file", None):
                    conn_dict["ansible_private_key_file"] = d.get("identity_file")
                if d.get("password", None):
                    conn_dict["ansible_password"] = d.get("password")
                    # Based on testinfra documentation, ansible password must be passed via ansible_ssh_pass
                    # issue to fix this can be found https://github.com/pytest-dev/pytest-testinfra/issues/580
                    conn_dict["ansible_ssh_pass"] = d.get("password")

                conn_dict["ansible_ssh_common_args"] = " ".join(
                    self.ssh_connection_options,
                )
            except StopIteration:
                return {}
            except OSError:
                # Instance has yet to be provisioned , therefore the
                # instance_config is not on disk.
                return {}
            else:
                return conn_dict

        return self.options.get("ansible_connection_options", {})

    def _created(self) -> str:
        if self.managed:
            return super()._created()
        return "unknown"

    def _get_instance_config(self, instance_name: str) -> dict[str, Any]:
        instance_config_dict = util.safe_load_file(self._config.driver.instance_config)

        return next(item for item in instance_config_dict if item["instance"] == instance_name)

    def sanity_checks(self) -> None:
        """Run sanity checks."""
        # Note(decentral1se): Cannot implement driver specifics are unknown

    def schema_file(self) -> str:
        """Return schema file path.

        Returns:
            Path to schema file.
        """
        return str(Path(data_module).parent / "driver.json")



#  Copyright (c) 2018 Marc Dequènes (Duck) <duck@redhat.com>  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import copy

import pytest

from molecule import config, scenarios


@pytest.fixture
def _instance(config_instance: config.Config):  # type: ignore[no-untyped-def]  # noqa: ANN202
    config_instance_1 = copy.deepcopy(config_instance)
    config_instance_1.config["scenario"]["name"] = "two"
    config_instance_1.molecule_file = config_instance_1.molecule_file.replace(
        "default",
        "02_foo",
    )

    config_instance_2 = copy.deepcopy(config_instance)
    config_instance_2.config["scenario"]["name"] = "one"
    config_instance_2.molecule_file = config_instance_2.molecule_file.replace(
        "default",
        "01_foo",
    )

    config_instance_3 = copy.deepcopy(config_instance)
    config_instance_3.config["scenario"]["name"] = "three"
    config_instance_3.molecule_file = config_instance_3.molecule_file.replace(
        "default",
        "03_foo",
    )

    return scenarios.Scenarios(
        [config_instance_1, config_instance_2, config_instance_3],
    )


def test_all_ordered(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    result = _instance.all

    assert len(result) == 3  # noqa: PLR2004
    assert result[0].name == "one"
    assert result[1].name == "two"
    assert result[2].name == "three"



"""Testinfra tests."""  # noqa: INP001

from __future__ import annotations

import os

import testinfra.utils.ansible_runner


testinfra_hosts = testinfra.utils.ansible_runner.AnsibleRunner(
    os.environ["MOLECULE_INVENTORY_FILE"],
).get_hosts("all")


def test_ansible_hostname(host):  # type: ignore[no-untyped-def]  # noqa: ANN201
    """Validate hostname."""
    f = host.file("/tmp/molecule/instance-1")  # noqa: S108

    assert not f.exists



# Copyright 2015 Docker, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Taken from Docker Compose:
# https://github.com/docker/compose/blob/master/compose/config/interpolation.py
"""Interpolation Module."""

from __future__ import annotations

import string

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Mapping, MutableMapping
    from re import Match


class InvalidInterpolation(Exception):  # noqa: N818
    """InvalidInterpolation Exception."""

    def __init__(
        self,
        string: str,  # pylint: disable=redefined-outer-name
        place: Exception,
    ) -> None:
        """Construct InvalidInterpolation.

        Args:
            string: Error message.
            place: The original exception with location information.
        """
        self.string = string
        self.place = place


class Interpolator:
    """Configuration options may contain environment variables.

    For example, suppose the shell contains ``VERIFIER_NAME=testinfra`` and
    the following molecule.yml is supplied.

    ```yaml
        verifier:
          - name: ${VERIFIER_NAME}
    ```

    Molecule will substitute ``$VERIFIER_NAME`` with the value of the
    ``VERIFIER_NAME`` environment variable.

    !!! warning

        If an environment variable is not set, Molecule substitutes with an
        empty string.

    Both ``$VARIABLE`` and ``${VARIABLE}`` syntax are supported. Extended
    shell-style features, such as ``${VARIABLE-default}`` and
    ``${VARIABLE:-default}`` are also supported. Even the default as another
    environment variable is supported like ``${VARIABLE-$DEFAULT}`` or
    ``${VARIABLE:-$DEFAULT}``. An empty string is returned when
    both variables are undefined.

    If a literal dollar sign is needed in a configuration, use a double dollar
    sign (`$$`).

    Molecule will substitute special ``MOLECULE_`` environment variables
    defined in `molecule.yml`.

    !!! note

        Remember, the ``MOLECULE_`` namespace is reserved for Molecule.  Do not
        prefix your own variables with `MOLECULE_`.

    A file may be placed in the root of the project as `.env.yml`, and Molecule
    will read variables when rendering `molecule.yml`.  See command usage.
    """

    def __init__(
        self,
        templater: type[TemplateWithDefaults],
        mapping: MutableMapping[str, str],
    ) -> None:
        """Construct Interpolator.

        Args:
            templater: A templater class.
            mapping: A mapping object.
        """
        self.templater = templater
        self.mapping = mapping

    def interpolate(
        self,
        string: str,  # pylint: disable=redefined-outer-name
        keep_string: str | None = None,
    ) -> str:
        """Template string with this Interpolator's mapping.

        Args:
            string: The string to modify.
            keep_string: A substring to keep intact, regardless of mapping.

        Returns:
            The modified string.

        Raises:
            InvalidInterpolation: When the interpolation fails.
        """
        try:
            return self.templater(string).substitute(self.mapping, keep_string)
        except ValueError as e:
            raise InvalidInterpolation(string, e) from e


class TemplateWithDefaults(string.Template):
    """TemplateWithDefaults Class.

    Attributes:
        idpattern: Replacement pattern for base Template class.
    """

    idpattern = r"[_a-z][_a-z0-9]*(?::?-[^}]+)?"

    # pylint: disable=too-many-return-statements  # pylint: disable=useless-suppression
    def substitute(  # type: ignore[override]  # pylint: disable=arguments-differ
        self,
        mapping: Mapping[str, str],
        keep_string: str | None,
    ) -> str:
        """Substitute substrings according to a mapping.

        Args:
            mapping: The mapping to apply to the string.
            keep_string: A substring to keep intact, regardless of mapping.

        Returns:
            The converted string.
        """

        # Helper function for .sub()
        def convert(mo: Match[str]) -> str | None:  # noqa: PLR0911
            # Check the most common path first.
            named = mo.group("named") or mo.group("braced")
            if named is not None:
                if keep_string and named.startswith(keep_string):
                    return f"${named}"
                if ":-" in named:
                    var, _, default = named.partition(":-")
                    # If default is also a variable
                    if default.startswith("$"):
                        default = mapping.get(default[1:], "")
                    return mapping.get(var) or default
                if "-" in named:
                    var, _, default = named.partition("-")
                    # If default is also a variable
                    if default.startswith("$"):
                        default = mapping.get(default[1:], "")
                    return mapping.get(var, default)
                val = mapping.get(named, "")
                return f"{val}"
            if mo.group("escaped") is not None:
                return self.delimiter
            if mo.group("invalid") is not None:
                self._invalid(mo)  # type: ignore[attr-defined]
                return None
            return None

        return self.pattern.sub(convert, self.template)  # type: ignore[arg-type]
